from typing import Final, List

import torch
import torch.nn.functional as F
from torch import Tensor, nn


def as_windowed(x: Tensor, window_length: int, step: int = 1, dim: int = 1) -> Tensor:
    """Returns a tensor with chunks of overlapping windows of the first dim of x.

    Args:
        x (Tensor): Input of shape [B, T, ...]
        window_length (int): Length of each window
        step (int): Step/hop of each window w.r.t. the original signal x
        dim (int): Dimension on to apply the windowing

    Returns:
        windowed tensor (Tensor): Output tensor with shape (if dim==1)
            [B, (N - window_length + step) // step, window_length, ...]
    """
    shape: List[int] = list(x.shape)
    stride: List[int] = list(x.stride())
    # stride: List[int] = [x.stride(i) for i in range(len(shape))]
    # shape[dim] = torch.div(shape[dim] - window_length + step, step, rounding_mode="trunc")
    shape[dim] = int(shape[dim] - window_length + step / step)
    if dim > 0:
        shape.insert(dim + 1, window_length)
        stride.insert(dim + 1, stride[dim])
    else:
        if dim == -1:
            shape.append(window_length)
            stride.append(stride[dim])
        else:
            shape.insert(dim, window_length)
            stride.insert(dim, stride[dim])
    stride[dim] = stride[dim] * step
    return x.as_strided(shape, stride)


class MultiFrameModule(nn.Module):
    """Multi-frame speech enhancement modules.

    Signal model and notation:
        Noisy: `x = s + n`
        Enhanced: `y = f(x)`
        Objective: `min ||s - y||`

        PSD: Power spectral density, notated eg. as `Rxx` for noisy PSD.
        IFC: Inter-frame correlation vector: PSD*u, u: selection vector. Notated as `rxx`
        RTF: Relative transfere function, also called steering vector.
    """

    num_freqs: Final[int]
    frame_size: Final[int]
    need_unfold: Final[bool]
    real: Final[bool]

    def __init__(self, num_freqs: int, frame_size: int, lookahead: int = 0, real: bool = False):
        """Multi-Frame filtering module.

        Args:
            num_freqs (int): Number of frequency bins used for filtering.
            frame_size (int): Frame size in FD domain.
            lookahead (int): Lookahead, may be used to select the output time step. Note: This
                module does not add additional padding according to lookahead!
        """
        super().__init__()
        self.num_freqs = num_freqs
        self.frame_size = frame_size
        self.real = real
        if real:
            self.pad = nn.ConstantPad3d((0, 0, 0, 0, frame_size - 1 - lookahead, lookahead), 0.0)
        else:
            self.pad = nn.ConstantPad2d((0, 0, frame_size - 1 - lookahead, lookahead), 0.0)
        self.need_unfold = frame_size > 1
        self.lookahead = lookahead

    def spec_unfold_real(self, spec: Tensor):
        if self.need_unfold:
            spec = self.pad(spec).unfold(-3, self.frame_size, 1)
            return spec.permute(0, 1, 5, 2, 3, 4)
            # return as_windowed(self.pad(spec), self.frame_size, 1, dim=-3)
        return spec.unsqueeze(-1)

    def spec_unfold(self, spec: Tensor):
        """Pads and unfolds the spectrogram according to frame_size.

        Args:
            spec (complex Tensor): Spectrogram of shape [B, C, T, F]
        Returns:
            spec (Tensor): Unfolded spectrogram of shape [B, C, T, F, N], where N: frame_size.
        """
        if self.need_unfold:
            return self.pad(spec).unfold(2, self.frame_size, 1)
        return spec.unsqueeze(-1)

    @staticmethod
    def solve(Rxx, rss, diag_eps: float = 1e-8, eps: float = 1e-7) -> Tensor:
        return torch.einsum(
            "...nm,...m->...n", torch.inverse(_tik_reg(Rxx, diag_eps, eps)), rss
        )  # [T, F, N]

    @staticmethod
    def apply_coefs(spec: Tensor, coefs: Tensor) -> Tensor:
        # spec: [B, C, T, F, N]
        # coefs: [B, C, T, F, N]
        return torch.einsum("...n,...n->...", spec, coefs)


def psd(x: Tensor, n: int) -> Tensor:
    """Compute the PSD correlation matrix Rxx for a spectrogram.

    That is, `X*conj(X)`, where `*` is the outer product.

    Args:
        x (complex Tensor): Spectrogram of shape [B, C, T, F]. Will be unfolded with `n` steps over
            the time axis.

    Returns:
        Rxx (complex Tensor): Correlation matrix of shape [B, C, T, F, N, N]
    """
    x = F.pad(x, (0, 0, n - 1, 0)).unfold(-2, n, 1)
    return torch.einsum("...n,...m->...mn", x, x.conj())


def df(spec: Tensor, coefs: Tensor) -> Tensor:
    """Deep filter implementation using `torch.einsum`. Requires unfolded spectrogram.

    Args:
        spec (complex Tensor): Spectrogram of shape [B, C, T, F, N]
        coefs (complex Tensor): Coefficients of shape [B, C, N, T, F]

    Returns:
        spec (complex Tensor): Spectrogram of shape [B, C, T, F]
    """
    return torch.einsum("...tfn,...ntf->...tf", spec, coefs)


def df_real(spec: Tensor, coefs: Tensor) -> Tensor:
    """Deep filter implementation for real valued input Tensors. Requires unfolded spectrograms.

    Args:
        spec (real-valued Tensor): Spectrogram of shape [B, C, N, T, F, 2].
        coefs (real-valued Tensor): Coefficients of shape [B, C, N, T, F, 2].

    Returns:
        spec (real-valued Tensor): Filtered Spectrogram of shape [B, C, T, F, 2]
    """
    b, c, _, t, f, _ = spec.shape
    out = torch.empty((b, c, t, f, 2), dtype=spec.dtype, device=spec.device)
    # real
    out[..., 0] = (spec[..., 0] * coefs[..., 0]).sum(dim=2)
    out[..., 0] -= (spec[..., 1] * coefs[..., 1]).sum(dim=2)
    # imag
    out[..., 1] = (spec[..., 0] * coefs[..., 1]).sum(dim=2)
    out[..., 1] += (spec[..., 1] * coefs[..., 0]).sum(dim=2)
    return out


class DF(MultiFrameModule):
    """Deep Filtering."""

    conj: Final[bool]

    def __init__(self, num_freqs: int, frame_size: int, lookahead: int = 0, conj: bool = False):
        super().__init__(num_freqs, frame_size, lookahead)
        self.conj = conj

    def forward(self, spec: Tensor, coefs: Tensor):
        spec_u = self.spec_unfold(torch.view_as_complex(spec))
        coefs = torch.view_as_complex(coefs)
        spec_f = spec_u.narrow(-2, 0, self.num_freqs)
        coefs = coefs.view(coefs.shape[0], -1, self.frame_size, coefs.shape[2], coefs.shape[3])
        if self.conj:
            coefs = coefs.conj()
        spec_f = df(spec_f, coefs)
        if self.training:
            spec = spec.clone()
        spec[..., : self.num_freqs, :] = torch.view_as_real(spec_f)
        return spec


class DFreal(MultiFrameModule):
    """Deep Filtering."""

    conj: Final[bool]

    def __init__(self, num_freqs: int, frame_size: int, lookahead: int = 0, conj: bool = False):
        super().__init__(num_freqs, frame_size, lookahead, real=True)
        self.conj = conj

    def forward(self, spec: Tensor, coefs: Tensor):
        """Pads and unfolds the spectrogram and applies deep filtering using only real valued types.

        Args:
            spec (Tensor): Spectrogram of shape [B, C, T, F, 2]
            coefs (Tensor): Spectrogram of shape [B, C, T, F, 2]
        """
        spec_u = self.spec_unfold_real(spec)
        spec_f = spec_u.narrow(-2, 0, self.num_freqs)
        new_shape = [coefs.shape[0], -1, self.frame_size] + list(coefs.shape[2:])
        coefs = coefs.view(new_shape)
        if self.conj:
            coefs = coefs.conj()
        spec_f = df_real(spec_f, coefs)
        spec[..., : self.num_freqs, :] = spec_f
        return spec


class CRM(MultiFrameModule):
    """Complex ratio mask."""

    def __init__(self, num_freqs: int, frame_size: int = 1, lookahead: int = 0):
        assert frame_size == 1 and lookahead == 0, (frame_size, lookahead)
        super().__init__(num_freqs, 1)

    def forward_impl(self, spec: Tensor, coefs: Tensor):
        return spec.squeeze(-1).mul(coefs)


class MfWf(MultiFrameModule):
    """Multi-frame Wiener filter base module."""

    cholesky_decomp: Final[bool]
    inverse: Final[bool]
    enforce_constraints: Final[bool]
    eps: Final[float]
    dload: Final[float]

    def __init__(
        self,
        num_freqs: int,
        frame_size: int,
        lookahead: int = 0,
        cholesky_decomp: bool = False,
        inverse: bool = True,
        enforce_constraints: bool = True,
        eps=1e-8,
        dload=1e-7,
    ):
        """Multi-frame Wiener Filter via an estimate of the inverse

        Args:
            num_freqs (int): Number of frequency bins to apply MVDR filtering to.
            frame_size (int): Frame size of the MF MVDR filter.
            lookahead (int): Lookahead of the frame.
            cholesky_decomp (bool): Whether the input is a cholesky decomposition of the correlation matrix. Defauls to `False`.
            inverse (bool): Whether the input is a normal or inverse correlation matrix. Defaults to `True`.
            enforce_constraints (bool): Enforce hermetian matrix for non-inverse input and a triangular matrix for cholesky decomposition inpiut.
        """
        super().__init__(num_freqs, frame_size, lookahead=lookahead)
        self.cholesky_decomp = cholesky_decomp
        self.inverse = inverse
        self.enforce_constraints = enforce_constraints
        self.triu_idcs = torch.triu_indices(self.frame_size, self.frame_size, 1)
        self.tril_idcs = torch.empty_like(self.triu_idcs)
        self.tril_idcs[0] = self.triu_idcs[1]
        self.tril_idcs[1] = self.triu_idcs[0]
        self.eps = eps
        self.dload = dload

    def get_r_factor(self):
        """Return an factor f such that Rxx/f in range [-1, 1]"""
        if self.inverse and self.cholesky_decomp:
            return 2e3
        elif self.inverse and not self.cholesky_decomp:
            return 3e7
        elif not self.inverse and self.cholesky_decomp:
            return 2e-4
        elif not self.inverse and not self.cholesky_decomp:
            return 5e-6

    def forward(self, spec: Tensor, ifc: Tensor, iRxx: Tensor) -> Tensor:
        """Multi-frame Wiener filter based on Rxx**-1 and speech IFC vector.

        Args:
            spec (complex Tensor): Spectrogram of shape [B, 1, T, F]
            ifc (complex Tensor): Inter-frame speech correlation vector [B, T, F, N*2]
            iRxx (complex Tensor): (Inverse) noisy covariance matrix Rxx**-1 [B, T, F, (N**2)*2] OR
                cholesky_decomp Rxx=L*L^H of the same shape.

        Returns:
            spec (complex Tensor): Filtered spectrogram of shape [B, C, T, F]
        """

        spec_u = self.spec_unfold(torch.view_as_complex(spec))
        iRxx = torch.view_as_complex(iRxx.unflatten(3, (self.frame_size, self.frame_size, 2)))
        if self.cholesky_decomp:
            if self.enforce_constraints:
                # Upper triangular (wo. diagonal) must be zero
                iRxx[:, :, :, self.triu_idcs[0], self.triu_idcs[1]] = 0.0
            # Revert cholesky decomposition
            iRxx = iRxx.matmul(iRxx.transpose(3, 4).conj())
        if self.enforce_constraints and not self.inverse and not self.cholesky_decomp:
            # If we have a cholesky_decomp input the constraints are already enforced.
            # We have a standard correlation matrix as input. Imaginary part on the diagonal should be 0.
            torch.diagonal(iRxx, dim1=-1, dim2=-2).imag = 0.0
            # Triu should be complex conj of tril
            tril_conj = iRxx[:, :, :, self.tril_idcs[0], self.tril_idcs[1]].conj()
            iRxx[:, :, :, self.triu_idcs[0], self.triu_idcs[1]] = tril_conj
        ifc = torch.view_as_complex(ifc.unflatten(3, (self.frame_size, 2)))
        spec_f = spec_u.narrow(-2, 0, self.num_freqs)
        if not self.inverse:  # Is already an inverse input. No need to inverse it again.
            # Regularization on diag for stability
            iRxx = _tik_reg(iRxx, self.dload, self.eps)
            # Compute weights by solving the equation system
            w = torch.linalg.solve(iRxx, ifc).unsqueeze(1)
        else:  # We already have an inverse estimate. Just compute the linear combination.
            w = torch.einsum("...nm,...m->...n", iRxx, ifc).unsqueeze(1)  # [B, 1, F, N]
        spec_f = self.apply_coefs(spec_f, w)
        if self.training:
            spec = spec.clone()
        spec[..., : self.num_freqs, :] = torch.view_as_real(spec_f)
        return spec


class MfMvdr(MultiFrameModule):
    """Multi-frame minimum variance distortionless beamformer based on Rnn**-1 and speech IFC vector."""

    cholesky_decomp: Final[bool]
    inverse: Final[bool]
    enforce_constraints: Final[bool]
    eps: Final[float]
    dload: Final[float]

    def __init__(
        self,
        num_freqs: int,
        frame_size: int,
        lookahead: int = 0,
        cholesky_decomp: bool = False,
        inverse: bool = True,
        enforce_constraints: bool = True,
        eps=1e-8,
        dload=1e-7,
    ):
        """Multi-frame minimum variance distortionless beamformer.

        Args:
            num_freqs (int): Number of frequency bins to apply MVDR filtering to.
            frame_size (int): Frame size of the MF MVDR filter.
            lookahead (int): Lookahead of the frame.
            cholesky_decomp (bool): Whether the input is a cholesky decomposition of the correlation matrix. Defauls to `False`.
            inverse (bool): Whether the input is a normal or inverse correlation matrix. Defaults to `True`.
            enforce_constraints (bool): Enforce hermetian matrix for non-inverse input and a triangular matrix for cholesky decomposition inpiut.
        """
        super().__init__(num_freqs, frame_size, lookahead=lookahead)
        self.cholesky_decomp = cholesky_decomp
        self.inverse = inverse
        self.enforce_constraints = enforce_constraints
        self.triu_idcs = torch.triu_indices(self.frame_size, self.frame_size, 1)
        self.tril_idcs = torch.empty_like(self.triu_idcs)
        self.tril_idcs[0] = self.triu_idcs[1]
        self.tril_idcs[1] = self.triu_idcs[0]
        self.eps = eps
        self.dload = dload

    def get_r_factor(self):
        """Return an factor f such that Rnn/f in range [-1, 1]"""
        if self.inverse and self.cholesky_decomp:
            return 2e4
        elif self.inverse and not self.cholesky_decomp:
            return 3e8
        elif not self.inverse and self.cholesky_decomp:
            return 5e-5
        elif not self.inverse and not self.cholesky_decomp:
            return 1e-6

    def forward(self, spec: Tensor, ifc: Tensor, iRnn: Tensor) -> Tensor:
        """Multi-frame MVDR filter based on Rnn**-1 and speech IFC vector.

        Args:
            spec (complex Tensor): Spectrogram of shape [B, C, T, F]
            ifc (complex Tensor): Inter-frame speech correlation vector [B, C*N*2, T, F]
            iRnn (complex Tensor): (Inverse) noise covariance matrix Rnn**-1 [B, T, F (N**2)*2] OR
                cholesky_decomp Rnn=L*L^H of the same shape.

        Returns:
            spec (complex Tensor): Filtered spectrogram of shape [B, C, T, F]
        """
        spec_u = self.spec_unfold(torch.view_as_complex(spec))
        iRnn = torch.view_as_complex(iRnn.unflatten(3, (self.frame_size, self.frame_size, 2)))
        if self.cholesky_decomp:
            if self.enforce_constraints:
                # Upper triangular (wo. diagonal) must be zero
                iRnn[:, :, :, self.triu_idcs[0], self.triu_idcs[1]] = 0.0
            # Revert cholesky decomposition
            iRnn = iRnn.matmul(iRnn.transpose(3, 4).conj())
        if self.enforce_constraints and not self.inverse and not self.cholesky_decomp:
            # If we have a cholesky_decomp input the constraints are already enforced.
            # We have a standard correlation matrix as input. Imaginary part on the diagonal should be 0.
            torch.diagonal(iRnn, dim1=-1, dim2=-2).imag = 0.0
            # Triu should be complex conj of tril
            tril_conj = iRnn[:, :, :, self.tril_idcs[0], self.tril_idcs[1]].conj()
            iRnn[:, :, :, self.triu_idcs[0], self.triu_idcs[1]] = tril_conj
        ifc = torch.view_as_complex(ifc.unflatten(3, (self.frame_size, 2)))
        spec_f = spec_u.narrow(-2, 0, self.num_freqs)
        if not self.inverse:  # Is already an inverse input. No need to inverse it again.
            # Regularization on diag for stability
            iRnn = _tik_reg(iRnn, self.dload, self.eps)
            # Compute weights by solving the equation system
            numerator = torch.linalg.solve(iRnn, ifc)
        else:  # We already have an inverse estimate. Just compute the linear combination.
            numerator = torch.einsum("...nm,...m->...n", iRnn, ifc)  # [B, C, F, N]
        denumerator = torch.einsum("...n,...n->...", ifc.conj(), numerator)
        # Normalize numerator
        scale = ifc[..., -1, None].conj()
        w = (numerator * scale / (denumerator.real.unsqueeze(-1) + self.eps)).unsqueeze(1)
        spec_f = self.apply_coefs(spec_f, w)
        if self.training:
            spec = spec.clone()
        spec[..., : self.num_freqs, :] = torch.view_as_real(spec_f)
        return spec


# From torchaudio
def _compute_mat_trace(input: torch.Tensor, dim1: int = -2, dim2: int = -1) -> torch.Tensor:
    r"""Compute the trace of a Tensor along ``dim1`` and ``dim2`` dimensions.
    Args:
        input (torch.Tensor): Tensor of dimension `(..., channel, channel)`
        dim1 (int, optional): the first dimension of the diagonal matrix
            (Default: -1)
        dim2 (int, optional): the second dimension of the diagonal matrix
            (Default: -2)
    Returns:
        Tensor: trace of the input Tensor
    """
    assert input.ndim >= 2, "The dimension of the tensor must be at least 2."
    assert (
        input.shape[dim1] == input.shape[dim2]
    ), "The size of ``dim1`` and ``dim2`` must be the same."
    input = torch.diagonal(input, 0, dim1=dim1, dim2=dim2)
    return input.sum(dim=-1)


def _tik_reg(mat: torch.Tensor, reg: float = 1e-7, eps: float = 1e-8) -> torch.Tensor:
    """Perform Tikhonov regularization (only modifying real part).
    Args:
        mat (torch.Tensor): input matrix (..., channel, channel)
        reg (float, optional): regularization factor (Default: 1e-8)
        eps (float, optional): a value to avoid the correlation matrix is all-zero (Default: ``1e-8``)
    Returns:
        Tensor: regularized matrix (..., channel, channel)
    """
    # Add eps
    C = mat.size(-1)
    eye = torch.eye(C, dtype=mat.dtype, device=mat.device)
    epsilon = _compute_mat_trace(mat).real[..., None, None] * reg
    # in case that correlation_matrix is all-zero
    epsilon = epsilon + eps
    mat = mat + epsilon * eye[..., :, :]
    return mat


def compute_corr(X: Tensor, N: int):
    Xw = F.pad(X, (0, 0, N - 1, 0)).unfold(1, N, 1)
    Rxx = torch.einsum("...n,...m->...mn", Xw, Xw.conj())
    return Rxx


def compute_ideal_wf(
    rxx_via_rssrnn=True, cholesky_decomp=False, inverse=True, enforce_constraints=True, manual=False
):
    from icecream import ic, install

    import libdf
    from df.config import config
    from df.io import load_audio, save_audio
    from df.model import ModelParams

    torch.set_printoptions(linewidth=140)

    ORDER = 5
    DLOAD = 1e-7
    EPS = 1e-8
    if cholesky_decomp and inverse:
        DLOAD = 1e-4
        EPS = 1e-5

    install()
    ic.includeContext = True

    config.use_defaults(allow_reload=True)
    p = ModelParams()
    p.fft_size = 96
    p.hop_size = 24
    p.sr = 24000
    n_freqs = p.fft_size // 2 + 1

    df = libdf.DF(sr=p.sr, fft_size=p.fft_size, hop_size=p.hop_size, nb_bands=p.nb_erb)
    s = load_audio("assets/clean_freesound_33711.wav", p.sr, num_frames=5 * p.sr)[0].mean(
        0, keepdim=True
    )
    n = load_audio("assets/noise_freesound_2530.wav", p.sr, num_frames=5 * p.sr)[0].mean(
        0, keepdim=True
    )
    x = s + n
    save_audio("out/noisy.wav", x, p.sr)

    wf = MfWf(
        n_freqs,
        ORDER,
        cholesky_decomp=cholesky_decomp,
        inverse=inverse,
        enforce_constraints=enforce_constraints,
    )

    X, S, N = [torch.from_numpy(df.analysis(x.numpy())) for x in (x, s, n)]
    Xw = F.pad(X, (0, 0, ORDER - 1, 0)).unfold(1, ORDER, 1)
    Rss, Rnn, Rxx = [compute_corr(A, ORDER) for A in (S, N, X)]
    ifc = Rss[..., -1]
    Rnn = _tik_reg(Rnn, DLOAD, EPS)
    if rxx_via_rssrnn:
        # Adding these is slightly better compared to estimating Rxx from X
        Rxx = Rss + Rnn
    else:
        Rxx = _tik_reg(Rxx, DLOAD, EPS)
    A = Rxx
    if inverse:
        A = torch.linalg.inv(A)
    if cholesky_decomp:
        A, info = torch.linalg.cholesky_ex(A)
        print("Number of errors during cholesky_decomp:", torch.where(info > 0, 1, 0).sum())
    ic(A.abs().mean())
    # Manual way
    if manual:
        w = torch.einsum("...nm,...m->...n", A, ifc)
        Y = torch.einsum("...fn,...fn->...f", Xw, w)
    else:
        # Using torch module (which expects real valued flattened input)
        Y = torch.view_as_complex(
            wf(
                torch.view_as_real(X).unsqueeze(1),
                torch.view_as_real(ifc).flatten(3),
                torch.view_as_real(A).flatten(3),
            ).squeeze(1)
        )
    y = df.synthesis(Y.numpy())
    save_audio("out/ideal_mfwf_c{}_i{}.wav".format(int(cholesky_decomp), int(inverse)), y, p.sr)


def compute_ideal_mvdr(cholesky_decomp=False, inverse=True, enforce_constraints=True, manual=False):
    from icecream import ic, install

    import libdf
    from df.config import config
    from df.io import load_audio, save_audio
    from df.model import ModelParams

    ic.includeContext = True
    torch.set_printoptions(linewidth=120)

    ORDER = 5
    DLOAD = 1e-7
    EPS = 1e-8
    if cholesky_decomp and inverse:
        DLOAD = 1e-4
        EPS = 1e-5

    install()

    config.use_defaults(allow_reload=True)
    p = ModelParams()
    p.fft_size = 96
    p.hop_size = 24
    p.sr = 24000
    n_freqs = p.fft_size // 2 + 1

    df = libdf.DF(sr=p.sr, fft_size=p.fft_size, hop_size=p.hop_size, nb_bands=p.nb_erb)
    s = load_audio("assets/clean_freesound_33711.wav", p.sr, num_frames=5 * p.sr)[0].mean(
        0, keepdim=True
    )
    n = load_audio("assets/noise_freesound_2530.wav", p.sr, num_frames=5 * p.sr)[0].mean(
        0, keepdim=True
    )
    x = s + n
    save_audio("out/noisy.wav", x, p.sr)

    mvdr = MfMvdr(
        n_freqs,
        ORDER,
        cholesky_decomp=cholesky_decomp,
        inverse=inverse,
        enforce_constraints=enforce_constraints,
    )

    X, S, N = [torch.from_numpy(df.analysis(x.numpy())) for x in (x, s, n)]
    Xw = F.pad(X, (0, 0, ORDER - 1, 0)).unfold(1, ORDER, 1)
    Rss, Rnn = [compute_corr(x, ORDER) for x in (S, N)]

    # A: IFC, needs to be normalized later
    ifc = Rss[..., -1]

    # B: IFC via EVD
    _, v = torch.linalg.eigh(Rss)
    ifc = v[..., -1]  # Choose highest eigenvector

    A = _tik_reg(Rnn, DLOAD, EPS)
    if inverse:
        A = torch.linalg.inv(A)
    if cholesky_decomp:
        A, info = torch.linalg.cholesky_ex(A)
        print("Number of errors during cholesky_decomp:", torch.where(info > 0, 1, 0).sum())
    ic(A.abs().mean())
    # Manual way
    if manual:
        ifc0 = ifc[..., -1]
        ifc = ifc / (ifc0.unsqueeze(-1) + EPS)
        if cholesky_decomp:
            A = A.matmul(A.conj().transpose(-1, -2))
        if inverse:
            num = torch.einsum("...nm,...m->...n", A, ifc)
        else:
            num = torch.linalg.solve(Rnn, ifc)
        denum = torch.einsum("...n,...n->...", ifc.conj(), num)
        w = num / (denum.unsqueeze(-1) + EPS)
        Y = torch.einsum("...fn,...fn->...f", Xw, w)
    # Using torch module (which expects real valued flattened input)
    else:
        Y = torch.view_as_complex(
            mvdr(
                torch.view_as_real(X).unsqueeze(1),
                torch.view_as_real(ifc).flatten(3),
                torch.view_as_real(A).flatten(3),
            ).squeeze(1)
        )
    y = df.synthesis(Y.numpy())
    save_audio("out/ideal_mfmvdr_c{}_i{}.wav".format(int(cholesky_decomp), int(inverse)), y, p.sr)


def compute_all_mf(enforce_constraints=True):
    for m in (compute_ideal_wf, compute_ideal_mvdr):
        for c in (True, False):
            for i in (True, False):
                m(cholesky_decomp=c, inverse=i, enforce_constraints=enforce_constraints)



from typing import List, Tuple

import torch

from df.enhance import init_df, setup_df_argument_parser
from df.logger import log_model_summary


def main(args):
    model, _, _ = init_df(
        args.model_base_dir,
        post_filter=args.pf,
        log_level=args.log_level,
        log_file=None,
        config_allow_defaults=True,
        epoch=None,
    )
    if args.type == "torch":
        print(model)
    elif args.type == "ptflops":
        log_model_summary(model, verbose=True, force=True)
    elif args.type == "table":
        for line in model_summary_table(model):
            print(line)
    else:
        raise NotImplementedError()


def model_summary_table(m: torch.nn.Module):
    def n_params(m: torch.nn.Module):
        return sum(p.numel() for p in m.parameters())

    childs = list(m.named_modules())[1:]
    max_len_name = max(len(n) for n, _ in childs) + 1
    out: List[Tuple[str, str]] = []
    # Heading
    out.append(("Name".ljust(max_len_name, " "), "Parameters"))
    out.append(("-" * len(out[0][0]), "-" * len(out[0][1])))
    # Add each module that has trainable parameters
    for name, c_m in childs:
        # Filter out parent modules that have children containing parameters
        if any(lambda c: n_params(c) > 0 for c in c_m.children()):
            continue
        if n_params(c_m) > 0:
            out.append((name.ljust(max_len_name, " "), f"{n_params(c_m):,}"))
    # Add sum of all parameters
    out.append(("-" * len(out[0][0]), "-" * len(out[0][1])))
    out.append(("Sum".ljust(max_len_name), f"{n_params(m):,}"))
    # Calculate max len of the parameters
    max_len_params = max(len(p) for _, p in out)
    # Left pad params and concat the strings
    for i in range(len(out)):
        out[i] = out[i][0] + out[i][1].rjust(max_len_params)
    return out


if __name__ == "__main__":
    parser = setup_df_argument_parser("INFO")
    parser.add_argument("--type", choices=["torch", "ptflops", "table"], default="torch")
    main(parser.parse_args())



import argparse
import os

import h5py
import torch

parser = argparse.ArgumentParser()
parser.add_argument("hdf5", nargs="+")
parser.add_argument("--keys", "-k", action="store_true")
parser.add_argument("--hours", action="store_true")
args = parser.parse_args()


def load_encoded(buffer, codec: str):
    import io

    import torchaudio as ta

    # In some rare cases, torch audio failes to fully decode vorbis resulting in a way shorter signal
    wav, _ = ta.load(io.BytesIO(buffer[...].tobytes()), format=codec.lower())
    return wav


total_hours = 0

for f in args.hdf5:
    assert os.path.isfile(f)
    with h5py.File(f, "r", libver="latest", swmr=True) as h5f:
        for n, k in h5f.attrs.items():
            print(f"Found attr {n} '{k}' in {f}")

        for group, samples in h5f.items():
            print(f"Found {len(samples)} samples in {group}")
            codec = h5f.attrs.get("codec", "pcm")
            sr = h5f.attrs["sr"]
            if args.keys or args.hours:
                for n, sample in samples.items():
                    if args.keys:
                        print(n)
                    if codec == "pcm":
                        audio = torch.from_numpy(sample[...])
                        if audio.dim() == 1:
                            audio.unsqueeze_(0)
                    else:
                        audio = load_encoded(sample, codec)
                    total_hours += audio.shape[-1] / sr / 60 / 60

if args.hours:
    print("Total len [h]:", total_hours)



import glob
import os

import torch
from loguru import logger

from df.enhance import init_df, save_audio, setup_df_argument_parser
from df.evaluation_utils import evaluation_loop_dns
from df.model import ModelParams


def main(args):
    model, df_state, suffix = init_df(
        args.model_base_dir,
        post_filter=args.pf,
        log_level=args.log_level,
        log_file="dnsmos-test.log",
        config_allow_defaults=True,
        epoch=args.epoch,
    )
    assert os.path.isdir(args.dataset_dir)
    sr = ModelParams().sr
    noisy_dir = args.dataset_dir
    noisy_files = glob.glob(noisy_dir + "/*.wav")
    n = len(noisy_files)
    logger.info(f"Evaluating DNSMOS on enhanced noisy files for dir: {noisy_dir} ({n} files)")
    if args.output_dir is not None:
        os.makedirs(args.output_dir, exist_ok=True)

    def save_audio_callback(cleanfn: str, enh):
        save_audio(
            os.path.basename(cleanfn),
            enh,
            sr,
            output_dir=args.output_dir,
            suffix=suffix,
            dtype=torch.float32,
        )

    dnsmos_versions = []
    if args.v1:
        dnsmos_versions += ["v1"]
    if args.v5 or len(dnsmos_versions) == 0:
        dnsmos_versions += ["v5"]
    metrics = evaluation_loop_dns(
        df_state,
        model,
        noisy_files,
        n_workers=args.metric_workers,
        save_audio_callback=save_audio_callback if args.output_dir is not None else None,
        metrics=["v5"],
        csv_path_enh=args.csv_path_enh,
        csv_path_noisy=args.csv_path_noisy,
        eval_noisy=False,
    )
    for k, v in metrics.items():
        logger.info(f"{k}: {v}")
    print(
        metrics["Enhanced SIG"],
        ",",
        metrics["Enhanced BAK"],
        ",",
        metrics["Enhanced OVRL"],
        ",",
        metrics["Enhanced P808_MOS"],
    )


if __name__ == "__main__":
    parser = setup_df_argument_parser()
    parser.add_argument(
        "dataset_dir",
        type=str,
        help="Test set directory containing noisy audio files.",
    )
    parser.add_argument("--v1", action="store_true")
    parser.add_argument("--v5", action="store_true")
    parser.add_argument(
        "--metric-workers",
        "-w",
        type=int,
        default=4,
        help="Number of worker processes for metric calculation.",
    )
    parser.add_argument(
        "--csv-path-enh",
        type=str,
        default=None,
        help="Path to csv score file containing metrics of enhanced audios.",
    )
    parser.add_argument(
        "--csv-path-noisy",
        type=str,
        default=None,
        help="Path to csv score file containing metrics of noisy audios.",
    )
    args = parser.parse_args()
    main(args)



import os
import shutil
import tarfile
from copy import deepcopy
from pathlib import Path
from typing import Dict, Iterable, List, Tuple, Union

import numpy as np
import onnx
import onnx.checker
import onnx.helper
import onnxruntime as ort
import torch
from loguru import logger
from torch import Tensor

from df.enhance import (
    ModelParams,
    df_features,
    enhance,
    get_model_basedir,
    init_df,
    setup_df_argument_parser,
)
from df.io import get_test_sample, save_audio
from libdf import DF


def shapes_dict(
    tensors: Tuple[Tensor], names: Union[Tuple[str], List[str]]
) -> Dict[str, Tuple[int]]:
    if len(tensors) != len(names):
        logger.warning(
            f"  Number of tensors ({len(tensors)}) does not match provided names: {names}"
        )
    return {k: v.shape for (k, v) in zip(names, tensors)}


def onnx_simplify(
    path: str, input_data: Dict[str, Tensor], input_shapes: Dict[str, Iterable[int]]
) -> str:
    import onnxsim

    model = onnx.load(path)
    model_simp, check = onnxsim.simplify(
        model,
        input_data=input_data,
        test_input_shapes=input_shapes,
    )
    model_n = os.path.splitext(os.path.basename(path))[0]
    assert check, "Simplified ONNX model could not be validated"
    logger.debug(model_n + ": " + onnx.helper.printable_graph(model.graph))
    try:
        onnx.checker.check_model(model_simp, full_check=True)
    except Exception as e:
        logger.error(f"Failed to simplify model {model_n}. Skipping: {e}")
        return path
    # new_path = os.path.join(os.path.dirname(path), model_n + "_simplified.onnx")
    onnx.save_model(model_simp, path)
    return path


def onnx_check(path: str, input_dict: Dict[str, Tensor], output_names: Tuple[str]):
    model = onnx.load(path)
    logger.debug(os.path.basename(path) + ": " + onnx.helper.printable_graph(model.graph))
    onnx.checker.check_model(model, full_check=True)
    sess = ort.InferenceSession(path, providers=["CPUExecutionProvider"])
    return sess.run(output_names, {k: v.numpy() for (k, v) in input_dict.items()})


def export_impl(
    path: str,
    model: torch.nn.Module,
    inputs: Tuple[Tensor, ...],
    input_names: List[str],
    output_names: List[str],
    dynamic_axes: Dict[str, Dict[int, str]],
    jit: bool = True,
    opset_version=14,
    check: bool = True,
    simplify: bool = True,
    print_graph: bool = False,
):
    export_dir = os.path.dirname(path)
    if not os.path.isdir(export_dir):
        logger.info(f"Creating export directory: {export_dir}")
        os.makedirs(export_dir)
    model_name = os.path.splitext(os.path.basename(path))[0]
    logger.info(f"Exporting model '{model_name}' to {export_dir}")

    input_shapes = shapes_dict(inputs, input_names)
    logger.info(f"  Input shapes: {input_shapes}")

    outputs = model(*inputs)
    output_shapes = shapes_dict(outputs, output_names)
    logger.info(f"  Output shapes: {output_shapes}")

    if jit:
        model = torch.jit.script(model, example_inputs=[tuple(a for a in inputs)])

    logger.info(f"  Dynamic axis: {dynamic_axes}")
    torch.onnx.export(
        model=deepcopy(model),
        f=path,
        args=inputs,
        input_names=input_names,
        dynamic_axes=dynamic_axes,
        output_names=output_names,
        opset_version=opset_version,
        keep_initializers_as_inputs=False,
    )

    input_dict = {k: v for (k, v) in zip(input_names, inputs)}
    if check:
        onnx_outputs = onnx_check(path, input_dict, tuple(output_names))
        for name, out, onnx_out in zip(output_names, outputs, onnx_outputs):
            try:
                np.testing.assert_allclose(
                    out.numpy().squeeze(), onnx_out.squeeze(), rtol=1e-6, atol=1e-5
                )
            except AssertionError as e:
                logger.warning(f"  Elements not close for {name}: {e}")
    if simplify:
        path = onnx_simplify(path, input_dict, shapes_dict(inputs, input_names))
        logger.info(f"  Saved simplified model {path}")
    if print_graph:
        onnx.helper.printable_graph(onnx.load_model(path).graph)

    return outputs


@torch.no_grad()
def export(
    model,
    export_dir: str,
    df_state: DF,
    check: bool = True,
    simplify: bool = True,
    opset=14,
    export_full: bool = False,
    print_graph: bool = False,
):
    model = deepcopy(model).to("cpu")
    model.eval()
    p = ModelParams()
    audio = torch.randn((1, 1 * p.sr))
    spec, feat_erb, feat_spec = df_features(audio, df_state, p.nb_df, device="cpu")

    # Export full model
    if export_full:
        path = os.path.join(export_dir, "deepfilternet2.onnx")
        input_names = ["spec", "feat_erb", "feat_spec"]
        dynamic_axes = {
            "spec": {2: "S"},
            "feat_erb": {2: "S"},
            "feat_spec": {2: "S"},
            "enh": {2: "S"},
            "m": {2: "S"},
            "lsnr": {1: "S"},
        }
        inputs = (spec, feat_erb, feat_spec)
        output_names = ["enh", "m", "lsnr", "coefs"]
        export_impl(
            path,
            model,
            inputs=inputs,
            input_names=input_names,
            output_names=output_names,
            dynamic_axes=dynamic_axes,
            jit=False,
            check=check,
            simplify=simplify,
            opset_version=opset,
            print_graph=print_graph,
        )

    # Export encoder
    feat_spec = feat_spec.transpose(1, 4).squeeze(4)  # re/im into channel axis
    path = os.path.join(export_dir, "enc.onnx")
    inputs = (feat_erb, feat_spec)
    input_names = ["feat_erb", "feat_spec"]
    dynamic_axes = {
        "feat_erb": {2: "S"},
        "feat_spec": {2: "S"},
        "e0": {2: "S"},
        "e1": {2: "S"},
        "e2": {2: "S"},
        "e3": {2: "S"},
        "emb": {1: "S"},
        "c0": {2: "S"},
        "lsnr": {1: "S"},
    }
    output_names = ["e0", "e1", "e2", "e3", "emb", "c0", "lsnr"]
    e0, e1, e2, e3, emb, c0, lsnr = export_impl(
        path,
        model.enc,
        inputs=inputs,
        input_names=input_names,
        output_names=output_names,
        dynamic_axes=dynamic_axes,
        jit=True,
        check=check,
        simplify=simplify,
        opset_version=opset,
        print_graph=print_graph,
    )
    np.savez_compressed(
        os.path.join(export_dir, "enc_input.npz"),
        feat_erb=feat_erb.numpy(),
        feat_spec=feat_spec.numpy(),
    )
    np.savez_compressed(
        os.path.join(export_dir, "enc_output.npz"),
        e0=e0.numpy(),
        e1=e1.numpy(),
        e2=e2.numpy(),
        e3=e3.numpy(),
        emb=emb.numpy(),
        c0=c0.numpy(),
        lsnr=lsnr.numpy(),
    )

    # Export erb decoder
    np.savez_compressed(
        os.path.join(export_dir, "erb_dec_input.npz"),
        emb=emb.numpy(),
        e0=e0.numpy(),
        e1=e1.numpy(),
        e2=e2.numpy(),
        e3=e3.numpy(),
    )
    inputs = (emb.clone(), e3, e2, e1, e0)
    input_names = ["emb", "e3", "e2", "e1", "e0"]
    output_names = ["m"]
    dynamic_axes = {
        "emb": {1: "S"},
        "e3": {2: "S"},
        "e2": {2: "S"},
        "e1": {2: "S"},
        "e0": {2: "S"},
        "m": {2: "S"},
    }
    path = os.path.join(export_dir, "erb_dec.onnx")
    m = export_impl(  # noqa
        path,
        model.erb_dec,
        inputs=inputs,
        input_names=input_names,
        output_names=output_names,
        dynamic_axes=dynamic_axes,
        jit=True,
        check=check,
        simplify=simplify,
        opset_version=opset,
        print_graph=print_graph,
    )
    np.savez_compressed(os.path.join(export_dir, "erb_dec_output.npz"), m=m.numpy())

    # Export df decoder
    np.savez_compressed(
        os.path.join(export_dir, "df_dec_input.npz"), emb=emb.numpy(), c0=c0.numpy()
    )
    inputs = (emb.clone(), c0)
    input_names = ["emb", "c0"]
    output_names = ["coefs"]
    dynamic_axes = {
        "emb": {1: "S"},
        "c0": {2: "S"},
        "coefs": {1: "S"},
    }
    path = os.path.join(export_dir, "df_dec.onnx")
    coefs = export_impl(  # noqa
        path,
        model.df_dec,
        inputs=inputs,
        input_names=input_names,
        output_names=output_names,
        dynamic_axes=dynamic_axes,
        jit=False,
        check=check,
        simplify=simplify,
        opset_version=opset,
        print_graph=print_graph,
    )
    np.savez_compressed(os.path.join(export_dir, "df_dec_output.npz"), coefs=coefs.numpy())


def main(args):
    try:
        import monkeytype  # noqa: F401
    except ImportError:
        print("Failed to import monkeytype. Please install it via")
        print("$ pip install MonkeyType")
        exit(1)

    print(args)
    model, df_state, _, epoch = init_df(
        args.model_base_dir,
        post_filter=args.pf,
        log_level=args.log_level,
        log_file="export.log",
        config_allow_defaults=True,
        epoch=args.epoch,
    )
    sample = get_test_sample(df_state.sr())
    enhanced = enhance(model, df_state, sample, True)
    out_dir = Path("out")
    if out_dir.is_dir():
        # attempt saving enhanced audio
        save_audio(os.path.join(out_dir, "enhanced.wav"), enhanced, df_state.sr())
    export_dir = Path(args.export_dir)
    export_dir.mkdir(parents=True, exist_ok=True)
    export(
        model,
        export_dir,
        df_state=df_state,
        opset=args.opset,
        check=args.check,
        simplify=args.simplify,
    )
    model_base_dir = get_model_basedir(args.model_base_dir)
    if model_base_dir != args.export_dir:
        shutil.copyfile(
            os.path.join(model_base_dir, "config.ini"),
            os.path.join(args.export_dir, "config.ini"),
        )
    model_name = Path(model_base_dir).name
    version_file = os.path.join(args.export_dir, "version.txt")
    with open(version_file, "w") as f:
        f.write(f"{model_name}_epoch_{epoch}")
    tar_name = export_dir / (Path(model_base_dir).name + "_onnx.tar.gz")
    with tarfile.open(tar_name, mode="w:gz") as f:
        f.add(os.path.join(args.export_dir, "enc.onnx"))
        f.add(os.path.join(args.export_dir, "erb_dec.onnx"))
        f.add(os.path.join(args.export_dir, "df_dec.onnx"))
        f.add(os.path.join(args.export_dir, "config.ini"))
        f.add(os.path.join(args.export_dir, "version.txt"))


if __name__ == "__main__":
    parser = setup_df_argument_parser()
    parser.add_argument("export_dir", help="Directory for exporting the onnx model.")
    parser.add_argument(
        "--no-check",
        help="Don't check models with onnx checker.",
        action="store_false",
        dest="check",
    )
    parser.add_argument("--simplify", help="Simply onnx models using onnxsim.", action="store_true")
    parser.add_argument("--opset", help="ONNX opset version", type=int, default=12)
    args = parser.parse_args()
    main(args)



import matplotlib.pyplot as plt
import numpy as np
from icecream import install  # noqa

from df.lr import cosine_scheduler

install()

LR = 1e-3
WARMUP = 3
LR_WARMUP_INIT = 1e-4
LR_MIN = 1e-6
CYCLE_DECAY = 1
CYCLE_LIM = -1
CYCLE_MUL = 1

NUM_EPOCHS = 100
STEPS_PER_EPOCH = 20


if __name__ == "__main__":
    lrs = cosine_scheduler(
        LR,
        LR_MIN,
        NUM_EPOCHS,
        STEPS_PER_EPOCH,
        WARMUP,
        LR_WARMUP_INIT,
        initial_ep_per_cycle=CYCLE_LIM,
        cycle_decay=CYCLE_DECAY,
        cycle_mul=CYCLE_MUL,
    )
    wds = cosine_scheduler(1e-12, 0.05, niter_per_ep=STEPS_PER_EPOCH, epochs=NUM_EPOCHS)
    t = np.linspace(0, NUM_EPOCHS, NUM_EPOCHS * STEPS_PER_EPOCH)
    bs = [8] + [16] + [24] * 3 + [32] * 5 + [64] * 10
    bs += [96] * (NUM_EPOCHS - len(bs))
    bs = np.repeat(bs, STEPS_PER_EPOCH)

    fig, ax = plt.subplots(figsize=(8, 3))
    fig.subplots_adjust(right=0.75)
    twin1 = ax.twinx()
    twin2 = ax.twinx()

    # Offset the right spine of twin2.  The ticks and label have already been
    # placed on the right by twinx above.
    twin2.spines.right.set_position(("axes", 1.12))

    (p1,) = ax.plot(t, lrs, "b-", label="Learning Rate")
    (p2,) = twin1.plot(t, wds, "r-", label="Weight Decay")
    (p3,) = twin2.plot(t, bs, "g-", label="Batch Size")

    ax.set_xlim(0, NUM_EPOCHS)
    ax.set_ylim(0, 0.0011)
    twin1.set_ylim(0, 0.055)
    twin2.set_ylim(0, 128)

    ax.set_xlabel("Epochs")
    ax.set_ylabel("Learning Rate")
    twin1.set_ylabel("Weight Decay")
    twin2.set_ylabel("Batch Size")

    ax.yaxis.label.set_color(p1.get_color())
    twin1.yaxis.label.set_color(p2.get_color())
    twin2.yaxis.label.set_color(p3.get_color())

    ax.grid()
    ax.set_xticks(range(0, NUM_EPOCHS, 5))
    twin2.set_yticks([0, 8, 16, 24, 32, 64, 96, 128])

    tkw = dict(size=4, width=1.5)
    ax.tick_params(axis="y", colors=p1.get_color(), **tkw)
    twin1.tick_params(axis="y", colors=p2.get_color(), **tkw)
    twin2.tick_params(axis="y", colors=p3.get_color(), **tkw)
    ax.tick_params(axis="x", **tkw)

    ax.legend(handles=[p1, p2, p3], loc="lower right")

    fig.set_tight_layout(True)

    plt.savefig("out/lrs.pdf")
    # plt.show()



# Usage:
# python dnsmos_local.py -t c:\temp\DNSChallenge4_Blindset -o DNSCh4_Blind.csv -p
#
import argparse
import concurrent.futures
import glob
import os
from time import sleep
from typing import List, Optional, Union

import librosa
import numpy as np
import pandas as pd
import soundfile as sf
import torch
from loguru import logger
from tqdm import tqdm

from df.io import load_audio, resample, save_audio
from df.logger import init_logger, log_metrics
from df.scripts.dnsmos import get_ort_session
from df.utils import download_file, get_cache_dir

SAMPLING_RATE = 16000
INPUT_LENGTH = 9.01
URL_ONNX = "https://github.com/microsoft/DNS-Challenge/raw/e14b010/DNSMOS/DNSMOS"
NAMES = ["SIG", "BAK", "OVRL", "P808_MOS"]
SLEEP_MS = int(os.environ.get("SLEEP", "0"))
if torch.cuda.is_available():
    TORCH_DEVICE = "cuda"
else:
    TORCH_DEVICE = "cpu"


class ComputeScore:
    def __init__(self, primary_model_path, p808_model_path, cpu: bool = False) -> None:
        self.onnx_sess = get_ort_session(primary_model_path, providers="cpu" if cpu else "gpu")
        self.p808_onnx_sess = get_ort_session(p808_model_path, providers="cpu" if cpu else "gpu")
        n_fft = 320
        self.w = torch.hann_window(n_fft + 1).to(TORCH_DEVICE)
        self.mel = torch.as_tensor(librosa.filters.mel(n_mels=120, sr=16000, n_fft=n_fft)).to(
            TORCH_DEVICE
        )

        # melspec: np.ndarray = np.einsum("...ft,mf->...mt", S, mel_basis, optimize=True)

    def audio_melspec_torch(
        self, audio, n_mels=120, frame_size=320, hop_length=160, sr=16000, to_db=True
    ):
        audio = torch.as_tensor(audio).to(TORCH_DEVICE).to(torch.float32)
        powspec = (
            torch.stft(
                audio,
                n_fft=frame_size + 1,
                hop_length=hop_length,
                window=self.w,
                return_complex=True,
            )
            .abs()
            .square()
        )
        mel_spec = torch.einsum("...ft,mf->mt", powspec, self.mel).t().cpu().numpy()
        if to_db:
            mel_spec = (librosa.power_to_db(mel_spec, ref=np.max) + 40) / 40
        return mel_spec

    def audio_melspec(
        self, audio, n_mels=120, frame_size=320, hop_length=160, sr=16000, to_db=True
    ):
        mel_spec = librosa.feature.melspectrogram(
            y=audio, sr=sr, n_fft=frame_size + 1, hop_length=hop_length, n_mels=n_mels
        )
        if to_db:
            mel_spec = (librosa.power_to_db(mel_spec, ref=np.max) + 40) / 40
        return mel_spec.T

    def get_polyfit_val(self, sig, bak, ovr, is_personalized_MOS):
        if is_personalized_MOS:
            p_ovr = np.poly1d([-0.00533021, 0.005101, 1.18058466, -0.11236046])
            p_sig = np.poly1d([-0.01019296, 0.02751166, 1.19576786, -0.24348726])
            p_bak = np.poly1d([-0.04976499, 0.44276479, -0.1644611, 0.96883132])
        else:
            p_ovr = np.poly1d([-0.06766283, 1.11546468, 0.04602535])
            p_sig = np.poly1d([-0.08397278, 1.22083953, 0.0052439])
            p_bak = np.poly1d([-0.13166888, 1.60915514, -0.39604546])

        sig_poly = p_sig(sig)
        bak_poly = p_bak(bak)
        ovr_poly = p_ovr(ovr)

        return sig_poly, bak_poly, ovr_poly

    def __call__(
        self,
        fpath_or_tensor: Union[str, torch.Tensor],
        sampling_rate: int,
        is_personalized_MOS: bool = False,
        fname: Optional[str] = None,
    ):
        fs = sampling_rate
        if isinstance(fpath_or_tensor, str):
            fpath = fpath_or_tensor
            logger.debug(f"Processing file: {fpath}")
            aud, input_fs = sf.read(fpath)
            if input_fs != fs:
                audio = librosa.resample(aud, orig_sr=input_fs, target_sr=fs)
            else:
                audio = aud
            feat_impl = self.audio_melspec
        else:
            audio = fpath_or_tensor
            feat_impl = self.audio_melspec_torch
            fpath = fname or "unknown"
        actual_audio_len = len(audio)
        len_samples = int(INPUT_LENGTH * fs)
        while len(audio) < len_samples:
            audio = np.append(audio, audio)

        num_hops = int(np.floor(len(audio) / fs) - INPUT_LENGTH) + 1
        hop_len_samples = fs
        predicted_mos_sig_seg_raw = []
        predicted_mos_bak_seg_raw = []
        predicted_mos_ovr_seg_raw = []
        predicted_mos_sig_seg = []
        predicted_mos_bak_seg = []
        predicted_mos_ovr_seg = []
        predicted_p808_mos = []

        for idx in range(num_hops):
            audio_seg = audio[
                int(idx * hop_len_samples) : int((idx + INPUT_LENGTH) * hop_len_samples)
            ]
            if len(audio_seg) < len_samples:
                continue

            input_features = np.array(audio_seg).astype("float32")[np.newaxis, :]
            p808_input_features = np.array(feat_impl(audio=audio_seg[:-160])).astype("float32")[
                np.newaxis, :, :
            ]
            oi = {"input_1": input_features}
            p808_oi = {"input_1": p808_input_features}
            p808_mos = self.p808_onnx_sess.run(None, p808_oi)[0][0][0]
            mos_sig_raw, mos_bak_raw, mos_ovr_raw = self.onnx_sess.run(None, oi)[0][0]
            mos_sig, mos_bak, mos_ovr = self.get_polyfit_val(
                mos_sig_raw, mos_bak_raw, mos_ovr_raw, is_personalized_MOS
            )
            predicted_mos_sig_seg_raw.append(mos_sig_raw)
            predicted_mos_bak_seg_raw.append(mos_bak_raw)
            predicted_mos_ovr_seg_raw.append(mos_ovr_raw)
            predicted_mos_sig_seg.append(mos_sig)
            predicted_mos_bak_seg.append(mos_bak)
            predicted_mos_ovr_seg.append(mos_ovr)
            predicted_p808_mos.append(p808_mos)

        clip_dict = {"filename": fname or fpath, "len_in_sec": actual_audio_len / fs, "sr": fs}
        clip_dict["num_hops"] = num_hops
        clip_dict["OVRL_raw"] = np.mean(predicted_mos_ovr_seg_raw)
        clip_dict["SIG_raw"] = np.mean(predicted_mos_sig_seg_raw)
        clip_dict["BAK_raw"] = np.mean(predicted_mos_bak_seg_raw)
        clip_dict["OVRL"] = np.mean(predicted_mos_ovr_seg)
        clip_dict["SIG"] = np.mean(predicted_mos_sig_seg)
        clip_dict["BAK"] = np.mean(predicted_mos_bak_seg)
        clip_dict["P808_MOS"] = np.mean(predicted_p808_mos)
        return clip_dict


def download_onnx_models():
    cache_dir = os.path.join(get_cache_dir(), "DNS5")
    if not os.path.isdir(cache_dir):
        os.makedirs(cache_dir)
    sig_bak_ovr = os.path.join(cache_dir, "sig_bak_ovr.onnx")
    if not os.path.exists(sig_bak_ovr):
        sig_bak_ovr = download_file(URL_ONNX + "/sig_bak_ovr.onnx", download_dir=cache_dir)
    p808 = os.path.join(cache_dir, "model_v8.onnx")
    if not os.path.exists(p808):
        p808 = download_file(URL_ONNX + "/model_v8.onnx", download_dir=cache_dir)
    return sig_bak_ovr, p808


def eval_sample_dnsmos(
    file: str,
    target_mos: Optional[List[float]] = None,
    log: bool = True,
    use_torch: bool = False,
    compute_score: Optional[ComputeScore] = None,
):
    primary_model_path, p808_model_path = download_onnx_models()
    compute_score = ComputeScore(primary_model_path, p808_model_path)
    if compute_score is None:
        primary_model_path, p808_model_path = download_onnx_models()
        compute_score = ComputeScore(primary_model_path, p808_model_path)
    desired_fs = SAMPLING_RATE
    if use_torch:
        audio = load_audio(file, desired_fs)[0].squeeze(0).numpy()
        scores = compute_score(audio, desired_fs, False, fname=file)
    scores = compute_score(file, desired_fs, False)
    scores = {n: scores[n] for n in NAMES}
    if log:
        logger.info(f"Processing file: {file}")
    if target_mos is not None:
        assert len(target_mos) == 4
        for n, t in zip(NAMES, target_mos):
            if not isclose(scores[n], t):
                diff = (np.asarray(target_mos) - np.fromiter(scores.values(), dtype=float)).tolist()
                log_metrics("Target   ", {n: v for (n, v) in zip(NAMES, target_mos)}, level="ERROR")
                log_metrics("Predicted", {n: v for (n, v) in scores.items()}, level="ERROR")
                log_metrics("Diff     ", {n: v for (n, v) in zip(NAMES, diff)}, level="ERROR")
                print(scores.values())
                exit(2)

    if log:
        log_metrics("Predicted", {n: v for (n, v) in scores.items()})
    return scores


def eval_dir_dnsmos(args):
    primary_model_path, p808_model_path = download_onnx_models()

    compute_score = ComputeScore(primary_model_path, p808_model_path)

    rows = []
    clips = []
    clips = glob.glob(os.path.join(args.testset_dir, "*.wav"))
    is_personalized_eval = False
    desired_fs = SAMPLING_RATE

    if len(clips) == 0:
        print(f"No samples found in dir {args.testset_dir}")
        exit(1)

    with concurrent.futures.ThreadPoolExecutor(max_workers=args.num_workers) as executor:
        future_to_url = {
            executor.submit(compute_score, clip, desired_fs, is_personalized_eval): clip
            for clip in clips
        }
        for future in tqdm(concurrent.futures.as_completed(future_to_url)):
            clip = future_to_url[future]
            try:
                data = future.result()
            except Exception as exc:
                print("%r generated an exception: %s" % (clip, exc))
                raise exc
            else:
                rows.append(data)

    df = pd.DataFrame(rows)
    if args.csv_file:
        csv_file = args.csv_file
        df.to_csv(csv_file)
    print_csv(df)


def load_encoded(buffer: np.ndarray, codec: str):
    import io

    import torchaudio as ta

    # In some rare cases, torch audio failes to fully decode vorbis resulting in a way shorter signal
    wav, _ = ta.load(io.BytesIO(buffer[...].tobytes()), format=codec.lower())
    return wav


def eval_ds(args):
    from tempfile import NamedTemporaryFile

    import h5py
    import torch

    primary_model_path, p808_model_path = download_onnx_models()
    compute_score = ComputeScore(primary_model_path, p808_model_path)

    rows = []
    is_personalized_eval = False
    desired_fs = SAMPLING_RATE

    def compute_score_audio(audio: torch.Tensor, sr: int, fname: str):
        if args.use_torch:
            audio = audio.squeeze(0)
            if audio.dtype == torch.int16:
                audio = audio.to(torch.float32) / 32767.0
            if sr != desired_fs:
                audio = resample(audio, sr, desired_fs)
            return compute_score(audio, desired_fs, is_personalized_eval, fname=fname)
        with NamedTemporaryFile(suffix=".wav") as nf:
            save_audio(nf.name, audio, sr, dtype=torch.float32)
            return compute_score(nf.name, desired_fs, is_personalized_eval, fname=fname)

    for path in args.ds:
        assert os.path.isfile(path)
        group = "speech"
        with h5py.File(path, "r", libver="latest") as f:
            print(f"Evaluating ds {path}")
            assert group in f
            sr = int(f.attrs["sr"])
            codec = f.attrs.get("codec", "pcm")
            for n, sample in f[group].items():  # type: ignore
                print(n)
                if codec == "pcm":
                    audio = torch.from_numpy(sample[...])
                    if audio.dim() == 1:
                        audio.unsqueeze_(0)
                else:
                    audio = load_encoded(sample, codec)
                if SLEEP_MS > 0:
                    sleep(SLEEP_MS / 1000)
                rows.append(compute_score_audio(audio, sr, fname=n))

    df = pd.DataFrame(rows)
    if args.csv_file:
        csv_file = args.csv_file
        df.to_csv(csv_file)
    print_csv(df)


def print_csv(df: Union[pd.DataFrame, List[str]]):
    if isinstance(df, list):
        df = [pd.read_csv(f) for f in df]
        df = pd.concat(df)
    print(df.describe())
    print(
        np.mean(df["SIG"]),
        np.mean(df["BAK"]),
        np.mean(df["OVRL"]),
        np.mean(df["P808_MOS"]),
    )


def isclose(a, b) -> bool:
    __a_tol = 1e-4
    __r_tol = 1e-4
    return abs(a - b) <= (__a_tol + __r_tol * abs(b))


if __name__ == "__main__":
    init_logger()
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="subparser_name")
    mn_parser = subparsers.add_parser("mean", aliases=["m"])
    mn_parser.add_argument("csv_file", type=str, nargs="+")
    eval_sample_parser = subparsers.add_parser("eval-sample")
    eval_sample_parser.add_argument("file", type=str)
    eval_sample_parser.add_argument("--target_mos", "-t", type=float, nargs="*")
    eval_dir_parser = subparsers.add_parser("eval-dir", aliases=["e"])
    eval_dir_parser.add_argument(
        "testset_dir", help="Path to the dir containing audio clips in .wav to be evaluated"
    )
    eval_dir_parser.add_argument(
        "-o", "--csv-file", help="If you want the scores in a CSV file provide the full path"
    )
    eval_dir_parser.add_argument("--cpu", help="Only run on CPU", action="store_true")
    eval_dir_parser.add_argument("--num-workers", type=int, default=1)
    eval_ds_parser = subparsers.add_parser("eval-ds")
    eval_ds_parser.add_argument("ds", help="Path to the hdf5 dataset file", nargs="+")
    eval_ds_parser.add_argument("--use-torch", action="store_true")
    eval_ds_parser.add_argument(
        "-o", "--csv-file", help="If you want the scores in a CSV file provide the full path"
    )

    args = parser.parse_args()
    if args.subparser_name is None:
        parser.print_help()
        exit(1)
    if args.subparser_name in ("m", "mean"):
        print_csv(args.csv_file)
    elif args.subparser_name == "eval-sample":
        eval_sample_dnsmos(args.file, args.target_mos)
    elif args.subparser_name == "eval-dir":
        eval_dir_dnsmos(args)
    elif args.subparser_name == "eval-ds":
        eval_ds(args)
    else:
        raise NotImplementedError()



import glob
import os

from loguru import logger
from torch import Tensor

from df.deepfilternet import ModelParams
from df.enhance import init_df, save_audio, setup_df_argument_parser
from df.evaluation_utils import evaluation_loop


def main(args):
    model, df_state, suffix = init_df(
        args.model_base_dir,
        post_filter=args.pf,
        log_level=args.log_level,
        log_file="voicebank-test.log",
        config_allow_defaults=True,
        epoch=args.epoch,
    )
    assert os.path.isdir(args.dataset_dir)
    sr = ModelParams().sr
    noisy_dir = os.path.join(args.dataset_dir, "noisy_testset_wav")
    clean_dir = os.path.join(args.dataset_dir, "clean_testset_wav")
    assert os.path.isdir(noisy_dir) and os.path.isdir(clean_dir)
    clean_files = sorted(glob.glob(clean_dir + "/*.wav"))
    noisy_files = sorted(glob.glob(noisy_dir + "/*.wav"))
    if args.output_dir is not None:
        logger.debug(f"Setting up output dir: {args.output_dir}")
        os.makedirs(args.output_dir, exist_ok=True)

    def save_audio_callback(cleanfn: str, enh: Tensor):
        save_audio(os.path.basename(cleanfn), enh, sr, output_dir=args.output_dir, suffix=suffix)

    metrics = evaluation_loop(
        df_state,
        model,
        clean_files,
        noisy_files,
        n_workers=args.metric_workers,
        save_audio_callback=save_audio_callback if args.output_dir is not None else None,
        metrics=["stoi", "composite", "sisdr"],
        csv_path_enh=args.csv_path_enh,
        csv_path_noisy=args.csv_path_noisy,
        noisy_metric=args.compute_noisy_metric,
        sleep_ms=args.sleep_ms,
    )
    for k, v in metrics.items():
        logger.info(f"{k}: {v}")
    print("".join(f"{m}," for k, m in metrics.items() if "SSNR" not in k)[:-1])


if __name__ == "__main__":
    parser = setup_df_argument_parser()
    parser.add_argument(
        "dataset_dir",
        type=str,
        help="Voicebank Demand Test set directory. Must contain 'noisy_testset_wav' and 'clean_testset_wav'.",
    )
    parser.add_argument(
        "--metric-workers",
        "-w",
        type=int,
        default=4,
        help="Number of worker processes for metric calculation.",
    )
    parser.add_argument(
        "--csv-path-enh",
        type=str,
        default=None,
        help="Path to csv score file containing metrics of enhanced audios.",
    )
    parser.add_argument(
        "--csv-path-noisy",
        type=str,
        default=None,
        help="Path to csv score file containing metrics of noisy audios.",
    )
    parser.add_argument("--compute-noisy-metric", action="store_true")
    parser.add_argument("--sleep-ms", type=int, default=0)
    args = parser.parse_args()
    main(args)



from typing import List, Tuple

import numpy as np
import torch
from loguru import logger
from torch import Tensor
from torch.nn import functional as F

from df.io import resample

EPS = np.finfo("float").eps


def as_windowed(x: Tensor, window_length: int, step: int = 1) -> Tensor:
    """Returns a tensor with chunks of overlapping windows of the first dim of x.

    Args:
        x (Tensor): Input of shape [N, B, H, W]
        window_length (int): Length of each window
        step (int): Step/hop of each window w.r.t. the original signal x

    Returns:
        windowed tensor (Tensor): Output tensor with shape
            [(N - window_length + step) // step, window_length, B, H, W]
    """
    shape = ((x.shape[0] - window_length + step) // step, window_length) + x.shape[1:]
    stride: List[int] = []
    for i in range(x.dim()):
        stride.append(x.stride(i))
    stride.insert(1, stride[0])
    stride[0] = stride[0] * step
    return x.as_strided(shape, stride)


def remove_silent_frames(
    x: Tensor, y: Tensor, dyn_range: int, framelen: int, hop: int, eps: float = EPS
) -> Tuple[List[Tensor], List[Tensor]]:
    """Remove the silent frames from each signal in the batch.

    Note:
        This implementation is based on https://github.com/mpariente/pystoi
        The overlap add code is based on https://github.com/pytorch/pytorchaudio

    Args:
        x (Tensor): Reference signal of shape [batch, samples].
        y (Tensor): Second signal of the same shape where the same frames are removed.
        dyn_range (int): Dynamic range / energy in [dB] to determin which frames to
            remove.
        framelen (int): Length for a frame that might be removed.
        hop (int): Hop between to subsequent frames.

    Returns:
        x (List[Tensor]): x without silent frames. Since each signal might have a
            different number of removed frames, the resulting signals will have
            different lengths. Thus a batch is return as list of tensors.
        y (List[Tensor]): y without silent frames.
    """
    pad = framelen - x.shape[1] % framelen
    pad_front, mod = divmod(pad, 2)
    pad_end = pad_front + mod
    x = F.pad(x, (pad_front, pad_end))
    y = F.pad(y, (pad_front, pad_end))
    B, _ = x.shape
    # [B, N] -> [N, B]
    x, y = x.t(), y.t()
    # Compute mask
    # Note: `framelen + 2` and `[1:-1]` is needed because otherwise the first sample is
    # 0 which results in NaN during overlapp add
    w = torch.hann_window(framelen + 2, periodic=False, device=x.device, dtype=x.dtype)
    w = w[1:-1].view(-1, 1)
    # Windowed signal length N_w = (N - L - H) // H; L=framelen, H=hop
    x_w = as_windowed(x, framelen, hop) * w  # [N_w, L, B]
    y_w = as_windowed(y, framelen, hop) * w  # [N_w, L, B]
    # Compute energies in dB
    # TODO: is `np.sqrt(win_size)` correct? Why not `w.sum().sqrt()` or torch.norm(w)?
    x_energies = 20 * torch.log10(x_w.norm(dim=1) / np.sqrt(framelen) + eps)  # [N_w, B]
    # Find boolean mask of energies lower than dynamic_range dB
    # with respect to maximum clean speech energy frame
    mask = (torch.max(x_energies, dim=0)[0] - dyn_range - x_energies).unsqueeze(1) < 0
    # Remove silent frames by masking for each sample in the batch
    x_w = [x_w[..., i].masked_select(mask[..., i]).view(-1, framelen) for i in range(B)]
    y_w = [y_w[..., i].masked_select(mask[..., i]).view(-1, framelen) for i in range(B)]
    n_no_sil_w = [x.shape[0] for x in x_w]
    # init zero arrays to hold x, y with silent frames removed
    n_no_sil = [(x.shape[0] - 1) * hop + framelen for x in x_w]
    x_no_sil = [torch.zeros((n_no_sil[i]), device=x.device) for i in range(B)]
    y_no_sil = [torch.zeros((n_no_sil[i]), device=x.device) for i in range(B)]
    # Overlapp add via transposed convolution
    x_w = [x.t().unsqueeze(0) for x in x_w]  # [N_w, L] -> [1, L, N_w]
    y_w = [y.t().unsqueeze(0) for y in y_w]  # [N_w, L] -> [1, L, N_w]
    eye = torch.eye(framelen, device=x.device, dtype=x.dtype).unsqueeze(1)
    # [1, L, N_w] -> [N]
    x_no_sil = [F.conv_transpose1d(x, eye, stride=hop).squeeze() for x in x_w]
    y_no_sil = [F.conv_transpose1d(y, eye, stride=hop).squeeze() for y in y_w]
    # Same for the window  [L, 1] -> [1, L, N_w]
    w = [w.repeat((1, n_no_sil_w[i])).unsqueeze(0) for i in range(B)]
    # [1, L, N_w] -> [N]
    w = [F.conv_transpose1d(w_, eye, stride=hop).squeeze() for w_ in w]
    x_no_sil = [x / w for x, w in zip(x_no_sil, w)]
    y_no_sil = [y / w for y, w in zip(y_no_sil, w)]
    # If the first frame is not masked out, we need to remove pad_front
    # Also maybe remove zero padding at the end
    for i in range(B):
        if mask[0, :, i]:
            x_no_sil[i] = x_no_sil[i][pad_front:]
            y_no_sil[i] = y_no_sil[i][pad_front:]
        if mask[-1, :, i]:
            x_no_sil[i] = x_no_sil[i][:-pad_end]
            y_no_sil[i] = y_no_sil[i][:-pad_end]
    return x_no_sil, y_no_sil


def thirdoct(fs, nfft, num_bands, min_freq):
    """Returns the 1/3 octave band matrix and its center frequencies
    # Arguments :
        fs : sampling rate
        nfft : FFT size
        num_bands : number of 1/3 octave bands
        min_freq : center frequency of the lowest 1/3 octave band
    # Returns :
        obm : Octave Band Matrix
        cf : center frequencies
    # Credit: https://github.com/mpariente/pystoi
    """
    f = np.linspace(0, fs, nfft + 1)
    f = f[: int(nfft / 2) + 1]
    k = np.array(range(num_bands)).astype(float)
    cf = np.power(2.0 ** (1.0 / 3), k) * min_freq
    freq_low = min_freq * np.power(2.0, (2 * k - 1) / 6)
    freq_high = min_freq * np.power(2.0, (2 * k + 1) / 6)
    obm = np.zeros((num_bands, len(f)))  # a verifier

    for i in range(len(cf)):
        # Match 1/3 oct band freq with fft frequency bin
        f_bin = np.argmin(np.square(f - freq_low[i]))
        freq_low[i] = f[f_bin]
        fl_ii = f_bin
        f_bin = np.argmin(np.square(f - freq_high[i]))
        freq_high[i] = f[f_bin]
        fh_ii = f_bin
        # Assign to the octave band matrix
        obm[i, fl_ii:fh_ii] = 1
    return obm, cf


def _stft(x, win_size, fft_size, hop_size, normalized=True, window=None):
    if window is None:
        ws = win_size + 2
        window = torch.hann_window(ws, periodic=False, device=x.device, dtype=x.dtype)[1:-1]
    # Pad the signal for (fft_size-win_size) / 2 at each side, because if win_size is <
    # fft_size, the first and last frames would not be considered for the non centered
    # (padded) version; this is inconsitent with scipy.signals stft
    missing_len = fft_size - win_size
    x = F.pad(x, (missing_len // 2, missing_len // 2))
    # To spectral domain
    spec = torch.stft(x, fft_size, hop_size, win_size, window, center=False, return_complex=False)
    # Normalize by default
    if normalized:
        spec /= window.sum().pow(2).sqrt()
    return spec


def stoi(x, y, fs_source):
    """Pytorch STOI implementation. Should only used for validation/developement, use pystoi for reporting test results.

    Arguments:
        x (Tensor): Target signal
        y (Tensor): Degraded signal
        fs_source (int): Sampling rate of input signals
    """
    assert x.shape == y.shape, "Inputs must have the same shape"
    assert x.dim() == 2, f"Expected input shape of [batch_size, samples], but got {x.shape}"
    fs = 10_000
    dyn_range = 40
    N_frame = 256
    N_fft = 512
    N_bands = 15
    min_freq = 150
    N = 30
    Beta = -15.0
    B = x.shape[0]  # batch size

    # Preallocate some stuff
    out = torch.empty(B, device=x.device, dtype=x.dtype)
    obm, _ = thirdoct(fs, N_fft, N_bands, min_freq)  # [N_fft//2-1, N_bands]
    obm = torch.from_numpy(obm).to(x)

    x = resample(x, fs_source, fs)
    y = resample(y, fs_source, fs)

    x_, y_ = remove_silent_frames(x, y, dyn_range, N_frame, N_frame // 2)

    for i in range(B):
        # To spectral domain
        if x_[i].numel() < N_fft:
            logger.warning("Could not calculate STOI (not enough frames left). Skipping.")
            continue
        x = _stft(x_[i], win_size=N_frame, fft_size=N_fft, hop_size=N_frame // 2)
        y = _stft(y_[i], win_size=N_frame, fft_size=N_fft, hop_size=N_frame // 2)
        # Power spectrogram
        x = x.pow(2).sum(-1)
        y = y.pow(2).sum(-1)
        # Reduce frequency res to 1/3 octave band
        x = torch.matmul(obm, x).sqrt()  # [N_bands, L]
        y = torch.matmul(obm, y).sqrt()
        if x.shape[-1] > N:
            x = x.unfold(-1, N, 1).permute(1, 2, 0)  # [L', N, N_bands]
            y = y.unfold(-1, N, 1).permute(1, 2, 0)
        else:
            # For short signals, we don't need the unfolding
            x = x.transpose(0, 1).unsqueeze(0)
            y = y.transpose(0, 1).unsqueeze(0)
        # Normalize per N-window
        norm = torch.norm(x, dim=1, keepdim=True) / (torch.norm(y, dim=1, keepdim=True) + EPS)
        y = y * norm
        # Clip
        c = 10 ** (-Beta / 20)
        y = torch.min(y, x * (1 + c))
        # Subtract mean vectors
        y = y - y.mean(dim=1, keepdim=True)
        x = x - x.mean(dim=1, keepdim=True)
        # Divide by norm
        x = x / (torch.norm(x, dim=1, keepdim=True) + EPS)
        y = y / (torch.norm(y, dim=1, keepdim=True) + EPS)
        corr = x * y
        # J, M as in eq. [6]
        J = x.shape[0]
        M = N_bands
        # Mean of all correlations
        out[i] = torch.sum(corr) / (M * J)
    return out



import argparse
import os
from typing import List, Tuple

import numpy as np
from torch import Tensor

from df.io import load_audio
from df.logger import init_logger, log_metrics
from df.scripts.dnsmos import SR, get_ort_session, isclose
from df.utils import download_file, get_cache_dir

URL_ONNX = "https://github.com/microsoft/DNS-Challenge/raw/82f1b17e7776a43eee395d0f45bae8abb700ad00/DNSMOS/DNSMOS/"
# Coefficients for polynomial fitting
P_SIG = np.poly1d([-0.08397278, 1.22083953, 0.0052439])
P_BAK = np.poly1d([-0.13166888, 1.60915514, -0.39604546])
P_OVR = np.poly1d([-0.06766283, 1.11546468, 0.04602535])
NAMES = ("SIG", "BAK", "OVL")
INPUT_LENGTH = 9.01


def main(args):
    file: str = args.file
    verbose = args.debug
    target_mos: List[float] = args.target_mos
    audio = load_audio(file, sr=SR, verbose=verbose)[0].squeeze(0)
    sig_bak_ovr = download_onnx_model()
    dnsmos = dnsmos_local(audio, sig_bak_ovr)
    log_metrics("Predicted", {n: v for (n, v) in zip(NAMES, dnsmos)})
    if target_mos is not None:
        if len(target_mos) > 0:
            assert len(target_mos) == len(dnsmos)
        log_metrics("Target   ", {n: v for (n, v) in zip(NAMES, target_mos)})
        for d, t in zip(dnsmos, target_mos):
            if not isclose(d, t):
                diff = (np.asarray(target_mos) - np.asarray(dnsmos)).tolist()
                log_metrics("Diff     ", {n: v for (n, v) in zip(NAMES, diff)}, level="ERROR")
                exit(2)
    exit(0)


def download_onnx_model():
    cache_dir = get_cache_dir()
    if not os.path.isdir(cache_dir):
        os.makedirs(cache_dir)
    name = "sig_bak_ovr.onnx"
    onnx = os.path.join(cache_dir, name)
    if not os.path.exists(onnx):
        onnx = download_file(URL_ONNX + name, download_dir=cache_dir)
    return onnx


def dnsmos_local(audio: Tensor, onnx: str) -> Tuple[float, float, float]:
    assert len(audio) >= SR, f"Audio to short: {audio.shape}"

    session = get_ort_session(onnx)

    if len(audio) < INPUT_LENGTH * SR:
        audio = np.pad(audio, (0, int(INPUT_LENGTH * SR - len(audio))), mode="wrap")
    num_hops = int(np.floor(len(audio) / SR) - INPUT_LENGTH) + 1
    hop_len_samples = SR
    predicted_mos_sig_seg = []
    predicted_mos_bak_seg = []
    predicted_mos_ovr_seg = []
    assert num_hops > 0

    for idx in range(num_hops):
        audio_seg = audio[int(idx * hop_len_samples) : int((idx + INPUT_LENGTH) * hop_len_samples)]
        if len(audio_seg) < INPUT_LENGTH * SR:
            continue
        input_features = np.array(audio_seg).astype("float32")[np.newaxis, :]
        oi = {"input_1": input_features}

        mos_sig_raw, mos_bak_raw, mos_ovr_raw = session.run(None, oi)[0][0]

        mos_sig = P_SIG(mos_sig_raw)
        mos_bak = P_BAK(mos_bak_raw)
        mos_ovr = P_BAK(mos_ovr_raw)

        predicted_mos_sig_seg.append(mos_sig)
        predicted_mos_bak_seg.append(mos_bak)
        predicted_mos_ovr_seg.append(mos_ovr)

    mod_sig = np.mean(predicted_mos_sig_seg)
    mod_bak = np.mean(predicted_mos_bak_seg)
    mod_ovr = np.mean(predicted_mos_ovr_seg)
    return mod_sig, mod_bak, mod_ovr


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--target-mos", "-t", type=float, nargs="*")
    parser.add_argument("--debug", "-d", "-v", action="store_true")
    parser.add_argument("file", type=str, help="Path to audio file for DNSMOS evaluation.")
    args = parser.parse_args()
    init_logger(level="DEBUG" if args.debug else "INFO")
    main(args)



#!/usr/bin/env python

import os
import unittest
from functools import partial
from typing import Callable, Dict, List, Optional, Union

import numpy as np
import torch
from loguru import logger
from torch import Tensor

import df
from df.enhance import DF, enhance, init_df
from df.evaluation_utils import HAS_OCTAVE, composite, si_sdr_speechmetrics, stoi
from df.io import load_audio, save_audio

__a_tol = 1e-4
__r_tol = 1e-4


def eval_metric(
    f: Callable[[np.ndarray, np.ndarray], np.ndarray],
    clean: Tensor,
    enhanced: Tensor,
    m_target,
    prefix: str = "",
    metric_name: str = "",
) -> bool:
    logger.info(prefix + f"Computing {metric_name} metrics")
    m_t = torch.as_tensor(m_target)  # target metric
    m_e = torch.as_tensor(
        f(clean.squeeze(0).numpy(), enhanced.squeeze(0).numpy())
    )  # enhanced metric
    m_e = m_e.to(torch.float32)
    logger.info(prefix + f"Expected {m_t}")
    logger.info(prefix + f"Got      {m_e}")
    is_close = torch.isclose(m_e, m_t, atol=__a_tol, rtol=__r_tol).all()
    if not is_close:
        logger.error(prefix + f"Diff     {m_t-m_e}")
    return is_close


TARGET_METRICS = {
    "DeepFilterNet": {
        "composite": [
            2.30728650093078,
            3.83064246177673,
            2.36408281326293,
            3.05453467369079,
            -2.7631254196166,
        ],
        "stoi": 0.9689496585281197,
        "sdr": 18.88543128967285,
    },
    "DeepFilterNet2": {
        "composite": [
            2.87284946441650,
            4.17169523239136,
            2.75626921653748,
            3.51172018051147,
            -0.91267710924149,
        ],
        "stoi": 0.9725977621169399,
        "sdr": 19.41733717918396,
    },
    "DeepFilterNet3": {
        "composite": [
            3.04712939262390,
            4.23114347457886,
            2.77058529853821,
            3.61812996864319,
            -1.51538455486298,
        ],
        "stoi": 0.9742409586906433,
        "sdr": 20.014915466308594,
    },
}


def _get_metric(name: str, sr: int):
    METRICS = {
        "composite": [partial(composite, sr=sr), partial(composite, sr=sr, use_octave=True)],
        "stoi": [partial(stoi, sr=sr)],
        "sdr": [si_sdr_speechmetrics],
    }
    return METRICS[name]


def _load_model(model_n: str, **kwargs):
    kwargs.setdefault("config_allow_defaults", True)
    model, df_state, _, epoch = init_df(model_n, **kwargs)
    logger.info(f"Loaded model {model_n} with epoch {epoch}")
    return model, df_state


class TestDfModels(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        torch.set_printoptions(precision=14, linewidth=120)
        os.makedirs("out", exist_ok=True)
        cls.df_dir = os.path.abspath(os.path.join(os.path.dirname(df.__file__), os.pardir))
        cls.models = {m: _load_model(m) for m in TARGET_METRICS.keys()}
        return cls

    def _test_model(
        self,
        model: torch.nn.Module,
        df_state: DF,
        target_metrics: Dict[str, Union[float, List[float]]],
        prefix: Optional[str] = None,
    ):
        out_n = f"out/enhanced_{prefix}.wav"
        prefix = prefix + " | " if prefix is not None else ""
        sr = df_state.sr()
        logger.info(prefix + "Loading audios")
        noisy, _ = load_audio(
            os.path.join(self.df_dir, os.path.pardir, "assets", "noisy_snr0.wav"), sr
        )
        clean, _ = load_audio(
            os.path.join(self.df_dir, os.path.pardir, "assets", "clean_freesound_33711.wav"), sr
        )
        enhanced = enhance(model, df_state, noisy, pad=True)
        save_audio(out_n, enhanced, sr)
        is_close = True
        for n in target_metrics.keys():
            for m in _get_metric(n, sr=sr):
                if not isinstance(m, partial) or "use_octave" not in m.keywords or HAS_OCTAVE:
                    cur_is_close = eval_metric(
                        m, clean, enhanced, m_target=target_metrics[n], prefix=prefix, metric_name=n
                    )
                    is_close = cur_is_close and is_close
        assert is_close

    def test_deepfilternet(self):
        model = "DeepFilterNet"
        self._test_model(*self.models[model], target_metrics=TARGET_METRICS[model], prefix=model)

    def test_deepfilternet2(self):
        model = "DeepFilterNet2"
        self._test_model(*self.models[model], target_metrics=TARGET_METRICS[model], prefix=model)

    def test_deepfilternet3(self):
        model = "DeepFilterNet3"
        self._test_model(*self.models[model], target_metrics=TARGET_METRICS[model], prefix=model)


if __name__ == "__main__":
    unittest.main()



import os
import re

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    # for Python <3.8 add 'importlib_metadata' as a dependency
    try:
        import importlib_metadata  # type: ignore
    except ModuleNotFoundError:
        importlib_metadata = None

VERSION_REGEX = r"\s*version\s*=\s*[\"']\s*([-.\w]{3,})\s*[\"']\s*"

try:
    if importlib_metadata is None:
        version = None
    else:
        version = importlib_metadata.version("deepfilternet")
except Exception:
    version = None
    path = os.path.join(os.path.dirname(__file__), os.pardir, "pyproject.toml")
    compiled_version_regex = re.compile(VERSION_REGEX)
    if os.path.isfile(path):
        for line in open(path):
            v_ = compiled_version_regex.search(line)
            if v_ is not None:
                version = v_.group(1).strip()
                break



import collections
import math
import os
import random
import subprocess
from socket import gethostname
from typing import Any, Optional, Set, Tuple, Union

import numpy as np
import torch
from loguru import logger
from torch import Tensor
from torch.autograd import Function
from torch.types import Number

from df.config import config
from df.model import ModelParams


def get_device():
    s = config("DEVICE", default="", section="train")
    if s == "":
        if torch.cuda.is_available():
            DEVICE = torch.device("cuda:0")
        else:
            DEVICE = torch.device("cpu")
    else:
        DEVICE = torch.device(s)
    return DEVICE


def as_complex(x: Tensor):
    if torch.is_complex(x):
        return x
    if x.shape[-1] != 2:
        raise ValueError(f"Last dimension need to be of length 2 (re + im), but got {x.shape}")
    if x.stride(-1) != 1:
        x = x.contiguous()
    return torch.view_as_complex(x)


def as_real(x: Tensor):
    if torch.is_complex(x):
        return torch.view_as_real(x)
    return x


class angle_re_im(Function):
    """Similar to torch.angle but robustify the gradient for zero magnitude."""

    @staticmethod
    def forward(ctx, re: Tensor, im: Tensor):
        ctx.save_for_backward(re, im)
        return torch.atan2(im, re)

    @staticmethod
    def backward(ctx, grad: Tensor) -> Tuple[Tensor, Tensor]:
        re, im = ctx.saved_tensors
        grad_inv = grad / (re.square() + im.square()).clamp_min_(1e-10)
        return -im * grad_inv, re * grad_inv


class angle(Function):
    """Similar to torch.angle but robustify the gradient for zero magnitude."""

    @staticmethod
    def forward(ctx, x: Tensor):
        ctx.save_for_backward(x)
        return torch.atan2(x.imag, x.real)

    @staticmethod
    def backward(ctx, grad: Tensor):
        (x,) = ctx.saved_tensors
        grad_inv = grad / (x.real.square() + x.imag.square()).clamp_min_(1e-10)
        return torch.view_as_complex(torch.stack((-x.imag * grad_inv, x.real * grad_inv), dim=-1))


def check_finite_module(obj, name="Module", _raise=True) -> Set[str]:
    out: Set[str] = set()
    if isinstance(obj, torch.nn.Module):
        for name, child in obj.named_children():
            out = out | check_finite_module(child, name)
        for name, param in obj.named_parameters():
            out = out | check_finite_module(param, name)
        for name, buf in obj.named_buffers():
            out = out | check_finite_module(buf, name)
    if _raise and len(out) > 0:
        raise ValueError(f"{name} not finite during checkpoint writing including: {out}")
    return out


def make_np(x: Union[Tensor, np.ndarray, Number]) -> np.ndarray:
    """Transforms Tensor to numpy.
    Args:
      x: An instance of torch tensor or caffe blob name

    Returns:
        numpy.array: Numpy array
    """
    if isinstance(x, np.ndarray):
        return x
    if np.isscalar(x):
        return np.array([x])
    if isinstance(x, Tensor):
        return x.detach().cpu().numpy()
    raise NotImplementedError(
        "Got {}, but numpy array, scalar, or torch tensor are expected.".format(type(x))
    )


def get_norm_alpha(log: bool = True) -> float:
    p = ModelParams()
    a_ = _calculate_norm_alpha(sr=p.sr, hop_size=p.hop_size, tau=p.norm_tau)
    precision = 3
    a = 1.0
    while a >= 1.0:
        a = round(a_, precision)
        precision += 1
    if log:
        logger.info(f"Running with normalization window alpha = '{a}'")
    return a


def _calculate_norm_alpha(sr: int, hop_size: int, tau: float):
    """Exponential decay factor alpha for a given tau (decay window size [s])."""
    dt = hop_size / sr
    return math.exp(-dt / tau)


def check_manual_seed(seed: Optional[int] = None):
    """If manual seed is not specified, choose a random one and communicate it to the user."""
    seed = seed or random.randint(1, 10000)
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    return seed


def get_git_root():
    """Returns the top level git directory or None if not called within the git repository."""
    try:
        git_local_dir = os.path.dirname(os.path.abspath(__file__))
        args = ["git", "-C", git_local_dir, "rev-parse", "--show-toplevel"]
        return subprocess.check_output(args).strip().decode()
    except subprocess.CalledProcessError:
        return None


def get_commit_hash():
    """Returns the current git commit."""
    try:
        git_dir = get_git_root()
        if git_dir is None:
            return None
        args = ["git", "-C", git_dir, "rev-parse", "--short", "--verify", "HEAD"]
        return subprocess.check_output(args).strip().decode()
    except subprocess.CalledProcessError:
        # probably not in git repo
        return None


def get_host() -> str:
    return gethostname()


def get_branch_name():
    try:
        git_dir = os.path.dirname(os.path.abspath(__file__))
        args = ["git", "-C", git_dir, "rev-parse", "--abbrev-ref", "HEAD"]
        branch = subprocess.check_output(args).strip().decode()
    except subprocess.CalledProcessError:
        # probably not in git repo
        branch = None
    return branch


# from pytorch/ignite:
def apply_to_tensor(input_, func):
    """Apply a function on a tensor or mapping, or sequence of tensors."""
    if isinstance(input_, torch.nn.Module):
        return [apply_to_tensor(c, func) for c in input_.children()]
    elif isinstance(input_, torch.nn.Parameter):
        return func(input_.data)
    elif isinstance(input_, Tensor):
        return func(input_)
    elif isinstance(input_, (str, bytes)):
        return input_
    elif isinstance(input_, collections.Mapping):
        return {k: apply_to_tensor(sample, func) for k, sample in input_.items()}
    elif isinstance(input_, collections.Iterable):
        return [apply_to_tensor(sample, func) for sample in input_]
    elif input_ is None:
        return input_
    else:
        return input_


def detach_hidden(hidden: Any) -> Any:
    """Cut backpropagation graph.
    Auxillary function to cut the backpropagation graph by detaching the hidden
    vector.
    """
    return apply_to_tensor(hidden, Tensor.detach)


def download_file(url: str, download_dir: str, extract: bool = False):
    import shutil
    import zipfile

    import requests

    local_filename = url.split("/")[-1]
    local_filename = os.path.join(download_dir, local_filename)
    with requests.get(url, stream=True) as r:
        if r.status_code >= 400:
            logger.error(f"Error downloading file {url} ({r.status_code}): {r.reason}")
            exit(1)
        with open(local_filename, "wb") as f:
            shutil.copyfileobj(r.raw, f)
    if extract:
        if os.path.splitext(local_filename)[1] != ".zip":
            logger.error("File not supported. Cannot extract.")
            exit(1)

        with zipfile.ZipFile(local_filename) as zf:
            zf.extractall(download_dir)
        os.remove(local_filename)

    return local_filename


def get_cache_dir():
    try:
        from appdirs import user_cache_dir

        return user_cache_dir("DeepFilterNet")
    except ImportError:
        import sys

        if sys.platform == "linux":
            return os.path.expanduser("~/.cache/DeepFilterNet/")
        else:
            raise ValueError(
                "Could not get cache dir. Please install `appdirs` via `pip install appdirs`"
            )



import glob
import os
import re

from loguru import logger

from df.enhance import init_df, save_audio, setup_df_argument_parser
from df.evaluation_utils import evaluation_loop
from df.model import ModelParams


def main(args):
    model, df_state, suffix = init_df(
        args.model_base_dir,
        post_filter=args.pf,
        log_level=args.log_level,
        log_file="dns2020-test.log",
        config_allow_defaults=True,
        epoch=args.epoch,
    )
    assert os.path.isdir(args.dataset_dir)
    sr = ModelParams().sr
    datasets = [os.path.join(args.dataset_dir, "no_reverb")]
    if args.reverb:
        datasets.append(os.path.join(args.dataset_dir, "with_reverb"))

    save_audio_callback = None  # type: ignore  # noqa: F811
    if args.output_dir is not None:
        os.makedirs(args.output_dir, exist_ok=True)

        def save_audio_callback(cleanfn: str, enh):  # noqa: F811
            save_audio(
                os.path.basename(cleanfn), enh, sr, output_dir=args.output_dir, suffix=suffix
            )

    for ds_dir in datasets:
        logger.info(f"Evaluating dataset {os.path.basename(ds_dir)}")
        assert os.path.isdir(ds_dir), ds_dir
        noisy_dir = os.path.join(ds_dir, "noisy")
        clean_dir = os.path.join(ds_dir, "clean")
        assert os.path.isdir(noisy_dir) and os.path.isdir(clean_dir)
        expr = re.compile(r"clnsp.*_fileid")
        noisy_files = glob.glob(noisy_dir + "/*.wav")
        clean_files = [
            re.sub(expr, "clean_fileid", f.replace("noisy", "clean")) for f in noisy_files
        ]
        assert len(clean_files) == 150

        metrics = evaluation_loop(
            df_state,
            model,
            clean_files,
            noisy_files,
            n_workers=args.metric_workers,
            save_audio_callback=save_audio_callback,
            metrics=["stoi", "composite", "sisdr"],
        )
        for k, v in metrics.items():
            logger.info(f"{k}: {v}")


if __name__ == "__main__":
    parser = setup_df_argument_parser()
    parser.add_argument(
        "dataset_dir",
        type=str,
        help="DNS 2020 Test set directory. Must contain 'no_reverb/clean' and 'no_reverb/noisy'.",
    )
    parser.add_argument(
        "--metric-workers",
        "-w",
        type=int,
        default=4,
        help="Number of worker processes for metric calculation.",
    )
    parser.add_argument(
        "--with-reverb", action="store_true", help="Also test on the reverb dataset", dest="reverb"
    )
    args = parser.parse_args()
    main(args)



import timeit
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import torch
from icecream import ic
from torch import Tensor

import libdf
from df.config import config
from df.io import get_test_sample, save_audio
from df.model import ModelParams
from df.utils import _calculate_norm_alpha
from df.visualization import spec_figure
from libdf import DF
from libdfdata import PytorchDataLoader as DataLoader

ic.includeContext = True


def rms(x) -> float:
    return torch.as_tensor(x).abs().pow(2).mean().sqrt().item()


def test_analysis_synthesis_stft(plot: Union[bool, str] = False):
    import matplotlib.pyplot as plt
    from icecream import ic, install

    install()

    sr = 24000
    n_fft = 96
    hop = n_fft // 4
    state = DF(sr=sr, fft_size=n_fft, hop_size=hop, nb_bands=32)
    sigin = get_test_sample(sr)
    ic(rms(sigin))

    # A note on normalization:
    # We have 3 options to normalize an FFT:
    #   - forward: Normalize by 1/N during forward path
    #   - backward: Normalize by 1/N during backward path
    #   - orthogonal: Normalize by sqrt(1/N) during both, forward and backward path
    #
    # DeepFilterNet does normalization during forward (i.e. stft) transform.
    # Unfortunately torch only offers two options:
    #   - normalized=True: Does orthogonal normalization
    #   - normalized=False: Does backward normalization
    #
    # To normalize an STFT, we also need to consider hop/overlap which slightly changes this
    # behaviour.
    #
    # To get the same normalization in torch, we will use the normalized = False, and manually
    # multiply by the normalization factor.
    # In the inverse transform, torch will apply the full normalization, thus we remove our
    # normalization again by dividing.
    stft_norm = 1 / (n_fft**2 / (2 * hop))
    w = torch.from_numpy(state.fft_window())
    freq_torch: Tensor = (
        torch.stft(
            sigin, n_fft=n_fft, hop_length=hop, window=w, return_complex=True, normalized=False
        ).transpose(1, 2)
        * stft_norm
    )
    sigout_torch: Tensor = torch.istft(
        freq_torch.transpose(1, 2) / stft_norm,
        n_fft=n_fft,
        hop_length=hop,
        window=w,
        normalized=False,
    )
    freq_rs = state.analysis(sigin.numpy())
    ic(rms(freq_rs))

    sigout_rs_torch: Tensor = torch.istft(
        torch.from_numpy(freq_rs).transpose(1, 2) / stft_norm,
        n_fft=n_fft,
        hop_length=hop,
        window=w,
        normalized=False,
    )
    ic(rms(sigout_torch))
    ic(rms(sigout_rs_torch))
    sigout_rs = state.synthesis(np.copy(freq_rs))
    ic(rms(sigout_rs))
    save_audio("out/fb_torch.wav", sigout_torch, sr)
    save_audio("out/fb_rs_torch.wav", sigout_rs_torch, sr)
    save_audio("out/fb_rs.wav", sigout_rs, sr)

    if isinstance(plot, str) or plot:
        target = freq_torch if plot == "torch" else torch.from_numpy(freq_rs)
        spec_figure(
            target.squeeze().T,
            sr,
            n_fft=n_fft,
            hop=hop,
            xlabel="time [s]",
            ylabel="freq",
            colorbar=True,
            ticks=True,
        )
        plt.show()


def plot_erb_spec():
    sr, fft, hop, nb_bands = 48000, 960, 480, 32
    df = DF(sr, fft, hop, nb_bands)
    x = get_test_sample(sr)
    x = df.analysis(x.reshape(1, -1).numpy())
    x = libdf.erb(x, df.erb_widths())
    spec_figure(x, sr)
    plt.ylim(0, 10000)
    plt.savefig("out_erb.pdf")


def test_timings(num=1000):
    import librosa  # noqa: F401
    import torch

    sr = 48000
    x = get_test_sample(sr).numpy()
    print("librosa: ", end="", flush=True)
    t = timeit.timeit(
        "librosa.stft(x, n_fft=960, hop_length=480, center=False)",
        globals=globals() | locals(),
        number=num,
    )
    print(f"{t:.3f}")

    sr, fft, hop, nb_bands = 48000, 960, 480, 32
    df = DF(sr, fft, hop, nb_bands)
    x = x.reshape(1, -1)
    print("dfrs: ", end="", flush=True)
    t = timeit.timeit("df.analysis(x)", globals=globals() | locals(), number=num)
    print(f"{t:.3f}")

    x = torch.from_numpy(x)
    w = torch.hann_window(960)
    print("Pytorch (with multi-threading): ", end="", flush=True)
    t = timeit.timeit(
        "torch.stft(x, n_fft=960, hop_length=480, window=w, center=False, return_complex=True)",
        globals=globals() | locals(),
        number=num,
    )
    print(f"{t:.3f}")


def test_dataloader(ds_dir):
    import os

    loader = DataLoader(
        ds_dir=os.path.expanduser(ds_dir),
        ds_config="../assets/dataset.cfg",
        sr=48_000,
        batch_size=8,
        split="train",
    )
    i = 0
    for batch in loader.iter_epoch(epoch=0):
        ic(i, batch)
        i += 1
        if i > 10:
            break
    ic("done")


def test_fft_dataloader(ds_dir):
    import os

    import librosa

    config.load("/tmp/dfrs1/config.ini")
    p = ModelParams()
    df = DF(p.sr, p.fft_size, p.hop_size, p.nb_erb)
    norm_alpha = _calculate_norm_alpha(p.sr, p.hop_size, tau=1)
    loader = DataLoader(
        ds_dir=os.path.expanduser(ds_dir),
        ds_config="assets/dataset.cfg",
        sr=p.sr,
        batch_size=1,
        fft_size=p.fft_size,
        hop_size=p.hop_size,
        num_workers=1,
        prefetch=1,
        nb_erb=p.nb_erb,
        nb_spec=p.nb_df,
        norm_alpha=norm_alpha,
    )
    for batch in loader.iter_epoch("train", seed=0):
        ic(batch.speech.shape)
        ic(batch.feat_spec.shape)  # [B, C, T, F]
        ic(batch.feat_erb.shape)
        ic(batch.atten)

        librosa.display.specshow(
            librosa.amplitude_to_db(np.abs(batch.speech[0].squeeze(0).T)),
            sr=p.sr,
            x_axis="time",
            y_axis="hz",
        )
        plt.savefig("out/spec.pdf")
        ic(
            batch.feat_spec.mean(),
            batch.feat_spec.var(),
            batch.feat_spec.abs().min(),
            batch.feat_spec.abs().max(),
        )
        librosa.display.specshow(
            batch.feat_erb[0].squeeze(0).numpy().T,
            sr=p.sr,
            x_axis="time",
            y_axis="log",
        )
        plt.colorbar()
        plt.savefig("out/erb.pdf")
        noisy_wav = df.synthesis(batch.noisy[0].numpy())
        clean_wav = df.synthesis(batch.speech[0].numpy())
        ic(noisy_wav.shape)
        save_audio("out/noisy.wav", noisy_wav.T, p.sr)
        save_audio("out/clean.wav", clean_wav.T, p.sr)
        break
    ic("done")



import numpy as np
from pesq import pesq
from scipy.linalg import toeplitz
from scipy.signal import stft

#################################################
#
#  SPEECH ENHANCEMENT PERFORMANCE METRICS
#
#  DENGFENG.KE @ 2020-03-14 BEIJING.CHINA
#
#  Modified version of https://raw.githubusercontent.com/IMLHF/PHASEN-PyTorch/9e68773/phasen_torch/sepm.py
#
#  This version is verified to be numerically close to the original octave/matlab version.


def extractOverlappedWindows(x, nperseg, noverlap, window=None):
    # source: https://github.com/scipy/scipy/blob/v1.2.1/scipy/signal/spectral.py
    step = nperseg - noverlap
    shape = x.shape[:-1] + ((x.shape[-1] - noverlap) // step, nperseg)
    strides = x.strides[:-1] + (step * x.strides[-1], x.strides[-1])
    result = np.lib.stride_tricks.as_strided(x, shape=shape, strides=strides)
    if window is not None:
        result = window * result
    return result


def SNRseg(clean_speech, processed_speech, fs, frameLen=0.03, overlap=0.75):
    eps = np.finfo(np.float64).eps

    winlength = round(frameLen * fs)  # window length in samples
    skiprate = int(np.floor((1 - overlap) * frameLen * fs))  # window skip in samples
    MIN_SNR = -10  # minimum SNR in dB
    MAX_SNR = 35  # maximum SNR in dB

    hannWin = 0.5 * (1 - np.cos(2 * np.pi * np.arange(1, winlength + 1) / (winlength + 1)))
    clean_speech_framed = extractOverlappedWindows(
        clean_speech, winlength, winlength - skiprate, hannWin
    )
    processed_speech_framed = extractOverlappedWindows(
        processed_speech, winlength, winlength - skiprate, hannWin
    )

    signal_energy = np.power(clean_speech_framed, 2).sum(-1)
    noise_energy = np.power(clean_speech_framed - processed_speech_framed, 2).sum(-1)

    segmental_snr = 10 * np.log10(signal_energy / (noise_energy + eps) + eps)
    segmental_snr[segmental_snr < MIN_SNR] = MIN_SNR
    segmental_snr[segmental_snr > MAX_SNR] = MAX_SNR
    segmental_snr = segmental_snr[:-1]  # remove last frame -> not valid
    return np.mean(segmental_snr)


def fwSNRseg(cleanSig, enhancedSig, fs, frameLen=0.03, overlap=0.75):
    if cleanSig.shape != enhancedSig.shape:
        raise ValueError("The two signals do not match!")
    eps = np.finfo(np.float64).eps
    cleanSig = cleanSig.astype(np.float64) + eps
    enhancedSig = enhancedSig.astype(np.float64) + eps
    winlength = round(frameLen * fs)  # window length in samples
    skiprate = int(np.floor((1 - overlap) * frameLen * fs))  # window skip in samples
    max_freq = fs / 2  # maximum bandwidth
    num_crit = 25  # number of critical bands
    n_fft = 2 ** np.ceil(np.log2(2 * winlength))
    n_fftby2 = int(n_fft / 2)
    gamma = 0.2

    cent_freq = np.zeros((num_crit,))
    bandwidth = np.zeros((num_crit,))

    cent_freq[0] = 50.0000
    bandwidth[0] = 70.0000
    cent_freq[1] = 120.000
    bandwidth[1] = 70.0000
    cent_freq[2] = 190.000
    bandwidth[2] = 70.0000
    cent_freq[3] = 260.000
    bandwidth[3] = 70.0000
    cent_freq[4] = 330.000
    bandwidth[4] = 70.0000
    cent_freq[5] = 400.000
    bandwidth[5] = 70.0000
    cent_freq[6] = 470.000
    bandwidth[6] = 70.0000
    cent_freq[7] = 540.000
    bandwidth[7] = 77.3724
    cent_freq[8] = 617.372
    bandwidth[8] = 86.0056
    cent_freq[9] = 703.378
    bandwidth[9] = 95.3398
    cent_freq[10] = 798.717
    bandwidth[10] = 105.411
    cent_freq[11] = 904.128
    bandwidth[11] = 116.256
    cent_freq[12] = 1020.38
    bandwidth[12] = 127.914
    cent_freq[13] = 1148.30
    bandwidth[13] = 140.423
    cent_freq[14] = 1288.72
    bandwidth[14] = 153.823
    cent_freq[15] = 1442.54
    bandwidth[15] = 168.154
    cent_freq[16] = 1610.70
    bandwidth[16] = 183.457
    cent_freq[17] = 1794.16
    bandwidth[17] = 199.776
    cent_freq[18] = 1993.93
    bandwidth[18] = 217.153
    cent_freq[19] = 2211.08
    bandwidth[19] = 235.631
    cent_freq[20] = 2446.71
    bandwidth[20] = 255.255
    cent_freq[21] = 2701.97
    bandwidth[21] = 276.072
    cent_freq[22] = 2978.04
    bandwidth[22] = 298.126
    cent_freq[23] = 3276.17
    bandwidth[23] = 321.465
    cent_freq[24] = 3597.63
    bandwidth[24] = 346.136

    bw_min = bandwidth[0]
    min_factor = np.exp(-30.0 / (2.0 * 2.303))  # % -30 dB point of filter

    all_f0 = np.zeros((num_crit,))
    crit_filter = np.zeros((num_crit, int(n_fftby2)))
    j = np.arange(0, n_fftby2)

    for i in range(num_crit):
        f0 = (cent_freq[i] / max_freq) * (n_fftby2)
        all_f0[i] = np.floor(f0)
        bw = (bandwidth[i] / max_freq) * (n_fftby2)
        norm_factor = np.log(bw_min) - np.log(bandwidth[i])
        crit_filter[i, :] = np.exp(-11 * (((j - np.floor(f0)) / bw) ** 2) + norm_factor)
        crit_filter[i, :] = crit_filter[i, :] * (crit_filter[i, :] > min_factor)

    num_frames = len(cleanSig) / skiprate - (winlength / skiprate)  # number of frames

    hannWin = 0.5 * (1 - np.cos(2 * np.pi * np.arange(1, winlength + 1) / (winlength + 1)))
    f, t, Zxx = stft(
        cleanSig[0 : int(num_frames) * skiprate + int(winlength - skiprate)],
        fs=fs,
        window=hannWin,
        nperseg=winlength,
        noverlap=winlength - skiprate,
        nfft=n_fft,
        detrend=False,
        return_onesided=True,
        boundary=None,
        padded=False,
    )
    clean_spec = np.abs(Zxx)
    clean_spec = clean_spec[:-1, :]
    clean_spec = clean_spec / clean_spec.sum(0)
    f, t, Zxx = stft(
        enhancedSig[0 : int(num_frames) * skiprate + int(winlength - skiprate)],
        fs=fs,
        window=hannWin,
        nperseg=winlength,
        noverlap=winlength - skiprate,
        nfft=n_fft,
        detrend=False,
        return_onesided=True,
        boundary=None,
        padded=False,
    )
    enh_spec = np.abs(Zxx)
    enh_spec = enh_spec[:-1, :]
    enh_spec = enh_spec / enh_spec.sum(0)

    clean_energy = crit_filter.dot(clean_spec)
    processed_energy = crit_filter.dot(enh_spec)
    error_energy = np.power(clean_energy - processed_energy, 2)
    error_energy[error_energy < eps] = eps
    W_freq = np.power(clean_energy, gamma)
    SNRlog = 10 * np.log10((clean_energy**2) / error_energy)
    fwSNR = np.sum(W_freq * SNRlog, 0) / np.sum(W_freq, 0)
    distortion = fwSNR.copy()
    distortion[distortion < -10] = -10
    distortion[distortion > 35] = 35

    return np.mean(distortion)


# def lpcoeff(speech_frame, model_order):
#    # ----------------------------------------------------------
#    # (1) Compute Autocorrelation Lags
#    # ----------------------------------------------------------

#     R=correlate(speech_frame,speech_frame)
#     R=R[len(speech_frame)-1:len(speech_frame)+model_order]
#    # ----------------------------------------------------------
#    # (2) Levinson-Durbin
#    # ----------------------------------------------------------
#     lpparams=np.ones((model_order+1))
#     lpparams[1:]=solve_toeplitz(R[0:-1],-R[1:])
#     return(lpparams,R)


def lpcoeff(speech_frame, model_order):
    # (1) Compute Autocor lags
    eps = np.finfo(np.float64).eps
    winlength = speech_frame.shape[0]
    R = []
    for k in range(model_order + 1):
        first = speech_frame[: (winlength - k)]
        second = speech_frame[k:winlength]
        R.append(np.sum(first * second))

    # (2) Lev-Durbin
    a = np.ones((model_order,))
    E = np.zeros((model_order + 1,))
    rcoeff = np.zeros((model_order,))
    E[0] = R[0]
    for i in range(model_order):
        if i == 0:
            sum_term = 0
        else:
            a_past = a[:i]
            sum_term = np.sum(a_past * np.array(R[i:0:-1]))
        # rcoeff[i] = (R[i+1] - sum_term)/E[i] # fixed by LiHongfeng
        rcoeff[i] = (R[i + 1] - sum_term) / max(E[i], eps)
        # if E[i] == 0:
        #   print(233333, i, eps==0)
        a[i] = rcoeff[i]
        if i > 0:
            a[:i] = a_past[:i] - rcoeff[i] * a_past[::-1]
        E[i + 1] = (1 - rcoeff[i] * rcoeff[i]) * E[i]
    acorr = np.array(R, dtype=np.float32)
    refcoeff = np.array(rcoeff, dtype=np.float32)
    a = a * -1
    lpparams = np.array([1] + list(a), dtype=np.float32)
    acorr = np.array(acorr, dtype=np.float32)
    refcoeff = np.array(refcoeff, dtype=np.float32)
    lpparams = np.array(lpparams, dtype=np.float32)

    # return acorr, refcoeff, lpparams
    return lpparams, acorr


def llr(clean_speech, processed_speech, fs, frameLen=0.03, overlap=0.75):
    eps = np.finfo(np.float64).eps
    alpha = 0.95
    winlength = round(frameLen * fs)  # window length in samples
    skiprate = int(np.floor((1 - overlap) * frameLen * fs))  # window skip in samples
    if fs < 10000:
        P = 10  # LPC Analysis Order
    else:
        P = 16  # this could vary depending on sampling frequency.

    hannWin = 0.5 * (1 - np.cos(2 * np.pi * np.arange(1, winlength + 1) / (winlength + 1)))
    clean_speech_framed = extractOverlappedWindows(
        clean_speech, winlength, winlength - skiprate, hannWin
    )
    processed_speech_framed = extractOverlappedWindows(
        processed_speech, winlength, winlength - skiprate, hannWin
    )
    numFrames = clean_speech_framed.shape[0]
    if numFrames == 0:
        return None
    numerators = np.zeros((numFrames - 1,))
    denominators = np.zeros((numFrames - 1,)) + eps

    for ii in range(numFrames - 1):
        A_clean, R_clean = lpcoeff(clean_speech_framed[ii, :], P)
        A_proc, R_proc = lpcoeff(processed_speech_framed[ii, :], P)

        numerators[ii] = A_proc.dot(toeplitz(R_clean).dot(A_proc.T))
        denominators[ii] = A_clean.dot(toeplitz(R_clean).dot(A_clean.T)) + eps

    frac = numerators / denominators
    frac[frac <= 0] = 1000
    distortion = np.log(frac)
    # distortion[distortion>2]=2 # this line is not in composite measure matlab implementation of loizou
    distortion = np.sort(distortion)
    distortion = distortion[: int(round(len(distortion) * alpha))]
    return np.mean(distortion)


def findLocPeaks(slope, energy):
    num_crit = len(energy)

    loc_peaks = np.zeros_like(slope)

    for ii in range(len(slope)):
        n = ii
        if slope[ii] > 0:
            while (n < num_crit - 1) and (slope[n] > 0):
                n = n + 1
            loc_peaks[ii] = energy[n - 1]
        else:
            while (n >= 0) and (slope[n] <= 0):
                n = n - 1
            loc_peaks[ii] = energy[n + 1]

    return loc_peaks


def wss(clean_speech, processed_speech, fs, frameLen=0.03, overlap=0.75):
    Kmax = 20  # value suggested by Klatt, pg 1280
    Klocmax = 1  # value suggested by Klatt, pg 1280
    alpha = 0.95
    if clean_speech.shape != processed_speech.shape:
        raise ValueError("The two signals do not match!")
    eps = np.finfo(np.float64).eps
    clean_speech = clean_speech.astype(np.float64) + eps
    processed_speech = processed_speech.astype(np.float64) + eps
    winlength = round(frameLen * fs)  # window length in samples
    skiprate = int(np.floor((1 - overlap) * frameLen * fs))  # window skip in samples
    max_freq = fs / 2  # maximum bandwidth
    num_crit = 25  # number of critical bands
    n_fft = 2 ** np.ceil(np.log2(2 * winlength))
    n_fftby2 = int(n_fft / 2)

    cent_freq = np.zeros((num_crit,))
    bandwidth = np.zeros((num_crit,))

    cent_freq[0] = 50.0000
    bandwidth[0] = 70.0000
    cent_freq[1] = 120.000
    bandwidth[1] = 70.0000
    cent_freq[2] = 190.000
    bandwidth[2] = 70.0000
    cent_freq[3] = 260.000
    bandwidth[3] = 70.0000
    cent_freq[4] = 330.000
    bandwidth[4] = 70.0000
    cent_freq[5] = 400.000
    bandwidth[5] = 70.0000
    cent_freq[6] = 470.000
    bandwidth[6] = 70.0000
    cent_freq[7] = 540.000
    bandwidth[7] = 77.3724
    cent_freq[8] = 617.372
    bandwidth[8] = 86.0056
    cent_freq[9] = 703.378
    bandwidth[9] = 95.3398
    cent_freq[10] = 798.717
    bandwidth[10] = 105.411
    cent_freq[11] = 904.128
    bandwidth[11] = 116.256
    cent_freq[12] = 1020.38
    bandwidth[12] = 127.914
    cent_freq[13] = 1148.30
    bandwidth[13] = 140.423
    cent_freq[14] = 1288.72
    bandwidth[14] = 153.823
    cent_freq[15] = 1442.54
    bandwidth[15] = 168.154
    cent_freq[16] = 1610.70
    bandwidth[16] = 183.457
    cent_freq[17] = 1794.16
    bandwidth[17] = 199.776
    cent_freq[18] = 1993.93
    bandwidth[18] = 217.153
    cent_freq[19] = 2211.08
    bandwidth[19] = 235.631
    cent_freq[20] = 2446.71
    bandwidth[20] = 255.255
    cent_freq[21] = 2701.97
    bandwidth[21] = 276.072
    cent_freq[22] = 2978.04
    bandwidth[22] = 298.126
    cent_freq[23] = 3276.17
    bandwidth[23] = 321.465
    cent_freq[24] = 3597.63
    bandwidth[24] = 346.136

    W = np.array(
        [
            0.003,
            0.003,
            0.003,
            0.007,
            0.010,
            0.016,
            0.016,
            0.017,
            0.017,
            0.022,
            0.027,
            0.028,
            0.030,
            0.032,
            0.034,
            0.035,
            0.037,
            0.036,
            0.036,
            0.033,
            0.030,
            0.029,
            0.027,
            0.026,
            0.026,
        ]
    )

    bw_min = bandwidth[0]
    min_factor = np.exp(-30.0 / (2.0 * 2.303))
    #      % -30 dB point of filter

    all_f0 = np.zeros((num_crit,))
    crit_filter = np.zeros((num_crit, int(n_fftby2)))
    j = np.arange(0, n_fftby2)

    for i in range(num_crit):
        f0 = (cent_freq[i] / max_freq) * (n_fftby2)
        all_f0[i] = np.floor(f0)
        bw = (bandwidth[i] / max_freq) * (n_fftby2)
        norm_factor = np.log(bw_min) - np.log(bandwidth[i])
        crit_filter[i, :] = np.exp(-11 * (((j - np.floor(f0)) / bw) ** 2) + norm_factor)
        crit_filter[i, :] = crit_filter[i, :] * (crit_filter[i, :] > min_factor)

    num_frames = len(clean_speech) / skiprate - (winlength / skiprate)  # number of frames

    hannWin = 0.5 * (1 - np.cos(2 * np.pi * np.arange(1, winlength + 1) / (winlength + 1)))
    scale = np.sqrt(1.0 / hannWin.sum() ** 2)

    f, t, Zxx = stft(
        clean_speech[0 : int(num_frames) * skiprate + int(winlength - skiprate)],
        fs=fs,
        window=hannWin,
        nperseg=winlength,
        noverlap=winlength - skiprate,
        nfft=n_fft,
        detrend=False,
        return_onesided=True,
        boundary=None,
        padded=False,
    )
    clean_spec = np.power(np.abs(Zxx) / scale, 2)
    clean_spec = clean_spec[:-1, :]

    f, t, Zxx = stft(
        processed_speech[0 : int(num_frames) * skiprate + int(winlength - skiprate)],
        fs=fs,
        window=hannWin,
        nperseg=winlength,
        noverlap=winlength - skiprate,
        nfft=n_fft,
        detrend=False,
        return_onesided=True,
        boundary=None,
        padded=False,
    )
    proc_spec = np.power(np.abs(Zxx) / scale, 2)
    proc_spec = proc_spec[:-1, :]

    clean_energy = crit_filter.dot(clean_spec)
    log_clean_energy = 10 * np.log10(clean_energy)
    log_clean_energy[log_clean_energy < -100] = -100
    proc_energy = crit_filter.dot(proc_spec)
    log_proc_energy = 10 * np.log10(proc_energy)
    log_proc_energy[log_proc_energy < -100] = -100

    log_clean_energy_slope = np.diff(log_clean_energy, axis=0)
    log_proc_energy_slope = np.diff(log_proc_energy, axis=0)

    dBMax_clean = np.max(log_clean_energy, axis=0)
    dBMax_processed = np.max(log_proc_energy, axis=0)

    numFrames = log_clean_energy_slope.shape[-1]

    clean_loc_peaks = np.zeros_like(log_clean_energy_slope)
    proc_loc_peaks = np.zeros_like(log_proc_energy_slope)
    for ii in range(numFrames):
        clean_loc_peaks[:, ii] = findLocPeaks(
            log_clean_energy_slope[:, ii], log_clean_energy[:, ii]
        )
        proc_loc_peaks[:, ii] = findLocPeaks(log_proc_energy_slope[:, ii], log_proc_energy[:, ii])

    Wmax_clean = Kmax / (Kmax + dBMax_clean - log_clean_energy[:-1, :])
    Wlocmax_clean = Klocmax / (Klocmax + clean_loc_peaks - log_clean_energy[:-1, :])
    W_clean = Wmax_clean * Wlocmax_clean

    Wmax_proc = Kmax / (Kmax + dBMax_processed - log_proc_energy[:-1])
    Wlocmax_proc = Klocmax / (Klocmax + proc_loc_peaks - log_proc_energy[:-1, :])
    W_proc = Wmax_proc * Wlocmax_proc

    W = (W_clean + W_proc) / 2.0

    distortion = np.sum(W * (log_clean_energy_slope - log_proc_energy_slope) ** 2, axis=0)
    distortion = distortion / np.sum(W, axis=0)
    distortion = np.sort(distortion)
    distortion = distortion[: int(round(len(distortion) * alpha))]
    return np.mean(distortion)


def composite(clean_speech, processed_speech, fs):
    """Computes composite measures: PESQ, CSIG, CBAK, COVL, SSNR."""
    assert fs == 16000
    wss_dist = wss(clean_speech, processed_speech, fs)
    llr_mean = llr(clean_speech, processed_speech, fs)
    if llr_mean is None:
        print("Error computing llr")
        llr_mean = 0
    segSNR = SNRseg(clean_speech, processed_speech, fs)
    pesq_mos = pesq(fs, clean_speech, processed_speech, "wb")

    Csig = 3.093 - 1.029 * llr_mean + 0.603 * pesq_mos - 0.009 * wss_dist
    Csig = np.max((1, Csig))
    Csig = np.min((5, Csig))  # limit values to [1, 5]
    Cbak = 1.634 + 0.478 * pesq_mos - 0.007 * wss_dist + 0.063 * segSNR
    Cbak = np.max((1, Cbak))
    Cbak = np.min((5, Cbak))  # limit values to [1, 5]
    Covl = 1.594 + 0.805 * pesq_mos - 0.512 * llr_mean - 0.007 * wss_dist
    Covl = np.max((1, Covl))
    Covl = np.min((5, Covl))  # limit values to [1, 5]
    return pesq_mos, Csig, Cbak, Covl, segSNR



import glob
import os
import re
from typing import List, Optional, Tuple, Union

import numpy as np
import torch
from loguru import logger
from torch import nn

from df.config import Csv, config
from df.model import init_model
from df.utils import check_finite_module
from libdf import DF


def get_epoch(cp) -> int:
    return int(os.path.basename(cp).split(".")[0].split("_")[-1])


def load_model(
    cp_dir: Optional[str],
    df_state: DF,
    jit: bool = False,
    mask_only: bool = False,
    train_df_only: bool = False,
    extension: str = "ckpt",
    epoch: Union[str, int, None] = "latest",
) -> Tuple[nn.Module, int]:
    if mask_only and train_df_only:
        raise ValueError("Only one of `mask_only` `train_df_only` can be enabled")
    model = init_model(df_state, run_df=mask_only is False, train_mask=train_df_only is False)
    if jit:
        model = torch.jit.script(model)
    blacklist: List[str] = config("CP_BLACKLIST", [], Csv(), save=False, section="train")  # type: ignore
    if cp_dir is not None:
        epoch = read_cp(
            model, "model", cp_dir, blacklist=blacklist, extension=extension, epoch=epoch
        )
        epoch = 0 if epoch is None else epoch
    else:
        epoch = 0
    return model, epoch


def read_cp(
    obj: Union[torch.optim.Optimizer, nn.Module],
    name: str,
    dirname: str,
    epoch: Union[str, int, None] = "latest",
    extension="ckpt",
    blacklist=[],
    log: bool = True,
):
    checkpoints = []
    if isinstance(epoch, str):
        assert epoch in ("best", "latest")
    if epoch == "best":
        checkpoints = glob.glob(os.path.join(dirname, f"{name}*.{extension}.best"))
        if len(checkpoints) == 0:
            logger.warning("Could not find `best` checkpoint. Checking for default...")
    if len(checkpoints) == 0:
        checkpoints = glob.glob(os.path.join(dirname, f"{name}*.{extension}"))
        checkpoints += glob.glob(os.path.join(dirname, f"{name}*.{extension}.best"))
    if len(checkpoints) == 0:
        return None
    if isinstance(epoch, int):
        latest = next((x for x in checkpoints if get_epoch(x) == epoch), None)
        if latest is None:
            logger.error(f"Could not find checkpoint of epoch {epoch}")
            exit(1)
    else:
        latest = max(checkpoints, key=get_epoch)
        epoch = get_epoch(latest)
    if log:
        logger.info("Found checkpoint {} with epoch {}".format(latest, epoch))
    latest = torch.load(latest, map_location="cpu",weights_only=False)
    #latest = torch.load(latest, map_location="cpu")
    latest = {k.replace("clc", "df"): v for k, v in latest.items()}
    if blacklist:
        reg = re.compile("".join(f"({b})|" for b in blacklist)[:-1])
        len_before = len(latest)
        latest = {k: v for k, v in latest.items() if reg.search(k) is None}
        if len(latest) < len_before:
            logger.info("Filtered checkpoint modules: {}".format(blacklist))
    if isinstance(obj, nn.Module):
        while True:
            try:
                missing, unexpected = obj.load_state_dict(latest, strict=False)
            except RuntimeError as e:
                e_str = str(e)
                logger.warning(e_str)
                if "size mismatch" in e_str:
                    latest = {k: v for k, v in latest.items() if k not in e_str}
                    continue
                raise e
            break
        for key in missing:
            logger.warning(f"Missing key: '{key}'")
        for key in unexpected:
            if key.endswith(".h0") or "erb_comp" in key:
                continue
            logger.warning(f"Unexpected key: {key}")
        return epoch
    obj.load_state_dict(latest)


def write_cp(
    obj: Union[torch.optim.Optimizer, nn.Module],
    name: str,
    dirname: str,
    epoch: int,
    extension="ckpt",
    metric: Optional[float] = None,
    cmp="min",
):
    check_finite_module(obj)
    n_keep = config("n_checkpoint_history", default=3, cast=int, section="train")
    n_keep_best = config("n_best_checkpoint_history", default=5, cast=int, section="train")
    if metric is not None:
        assert cmp in ("min", "max")
        metric = float(metric)  # Make sure it is not an integer
        # Each line contains a previous best with entries: (epoch, metric)
        with open(os.path.join(dirname, ".best"), "a+") as prev_best_f:
            prev_best_f.seek(0)  # "a+" creates a file in read/write mode without truncating
            lines = prev_best_f.readlines()
            if len(lines) == 0:
                prev_best = float("inf" if cmp == "min" else "-inf")
            else:
                prev_best = float(lines[-1].strip().split(" ")[1])
            cmp = "__lt__" if cmp == "min" else "__gt__"
            if getattr(metric, cmp)(prev_best):
                logger.info(f"Saving new best checkpoint at epoch {epoch} with metric: {metric}")
                prev_best_f.seek(0, os.SEEK_END)
                np.savetxt(prev_best_f, np.array([[float(epoch), metric]]))
                cp_name = os.path.join(dirname, f"{name}_{epoch}.{extension}.best")
                torch.save(obj.state_dict(), cp_name)
                cleanup(name, dirname, extension + ".best", nkeep=n_keep_best)
    cp_name = os.path.join(dirname, f"{name}_{epoch}.{extension}")
    logger.info(f"Writing checkpoint {cp_name} with epoch {epoch}")
    torch.save(obj.state_dict(), cp_name)
    cleanup(name, dirname, extension, nkeep=n_keep)


def cleanup(name: str, dirname: str, extension: str, nkeep=5):
    if nkeep < 0:
        return
    checkpoints = glob.glob(os.path.join(dirname, f"{name}*.{extension}"))
    if len(checkpoints) == 0:
        return
    checkpoints = sorted(checkpoints, key=get_epoch, reverse=True)
    for cp in checkpoints[nkeep:]:
        logger.debug("Removing old checkpoint: {}".format(cp))
        os.remove(cp)


def check_patience(
    dirname: str, max_patience: int, new_metric: float, cmp: str = "min", raise_: bool = True
):
    cmp = "__lt__" if cmp == "min" else "__gt__"
    new_metric = float(new_metric)  # Make sure it is not an integer
    prev_patience, prev_metric = read_patience(dirname)
    if prev_patience is None or getattr(new_metric, cmp)(prev_metric):
        # We have a better new_metric, reset patience
        write_patience(dirname, 0, new_metric)
    else:
        # We don't have a better metric, decrement patience
        new_patience = prev_patience + 1
        write_patience(dirname, new_patience, prev_metric)
        if new_patience >= max_patience:
            msg = f"No improvements on validation metric ({prev_metric:.3f} - {new_metric:.3f}) for {max_patience} epochs. Stopping."
            if raise_:
                raise ValueError(msg)
            else:
                logger.warning(msg)
                return False
    return True


def read_patience(dirname: str) -> Tuple[Optional[int], float]:
    fn = os.path.join(dirname, ".patience")
    if not os.path.isfile(fn):
        return None, 0.0
    patience, metric = np.loadtxt(fn)
    return int(patience), float(metric)


def write_patience(dirname: str, new_patience: int, metric: float):
    return np.savetxt(os.path.join(dirname, ".patience"), [new_patience, metric])


def test_check_patience():
    import tempfile

    with tempfile.TemporaryDirectory() as d:
        check_patience(d, 3, 1.0)
        check_patience(d, 3, 1.0)
        check_patience(d, 3, 1.0)
        assert check_patience(d, 3, 1.0, raise_=False) is False

    with tempfile.TemporaryDirectory() as d:
        check_patience(d, 3, 1.0)
        check_patience(d, 3, 0.9)
        check_patience(d, 3, 1.0)
        check_patience(d, 3, 1.0)
        assert check_patience(d, 3, 1.0, raise_=False) is False

    with tempfile.TemporaryDirectory() as d:
        check_patience(d, 3, 1.0, cmp="max")
        check_patience(d, 3, 1.9, cmp="max")
        check_patience(d, 3, 1.0, cmp="max")
        check_patience(d, 3, 1.0, cmp="max")
        assert check_patience(d, 3, 1.0, cmp="max", raise_=False) is False



import os
import sys

import torchaudio as ta

from df.visualization import plt, spec_figure


def main(path: str):
    audio, sr = ta.load(path)
    f = plt.figure(figsize=(10, 6))
    f.set_tight_layout(True)
    spec_figure(
        audio,
        sr,
        colorbar=True,
        from_audio=True,
        figure=f,
        labels=True,
        colorbar_format="%+2.0f dB",
        cmap="inferno",
        n_fft=2048,
        hop=512,
        kHz=True,
    )
    plt.savefig(os.path.splitext(path)[0] + ".pdf")


if __name__ == "__main__":
    assert len(sys.argv) == 2
    assert os.path.isfile(sys.argv[1])
    main(sys.argv[1])



from .config import config
from .enhance import enhance, init_df
from .version import version
from .train import run_df_train  # 添加 run_df_train 函数

# 预加载配置文件
try:
    import os
    # 首先检查环境变量
    if "DF_CONFIG_PATH" in os.environ and os.path.isfile(os.environ["DF_CONFIG_PATH"]):
        config.load(os.environ["DF_CONFIG_PATH"], allow_defaults=True)
    else:
        # 然后尝试Docker环境和其他常见位置
        possible_paths = [
            "/workspace/AudioClarity/models/DeepFilterNet2/config.ini",
            os.path.join(os.getcwd(), "models/DeepFilterNet2/config.ini"),
            os.path.join(os.getcwd(), "config.ini"),
            "/AudioClarity/models/DeepFilterNet2/config.ini"
        ]
        for p in possible_paths:
            if os.path.isfile(p):
                config.load(p, allow_defaults=True)
                break
except Exception as e:
    # 静默失败，不打印错误消息，以避免干扰正常导入
    pass

# 更新 __all__ 列表，包含 run_df_train
__all__ = ["config", "version", "enhance", "init_df", "run_df_train"]
__version__ = version


import os
import sys

from df.enhance import init_df


def main():
    dir = sys.argv[1] if len(sys.argv) > 1 else None
    if dir is not None:
        assert os.path.isdir(dir), f"Model base directory not found: {dir}"
    model = init_df(dir, log_level="ERROR", config_allow_defaults=True)[0]
    print(model)


if __name__ == "__main__":
    main()
